@@ -208,7 +208,7 @@ defmodule Exception do
          {_, kind, _, clauses} <- List.keyfind(defs, {function, arity}, 0) do
       clauses =
         for {meta, ex_args, guards, _block} <- clauses do
-          scope = :elixir_erl.scope(meta)
+          scope = :elixir_erl.scope(meta, true)
 
           {erl_args, scope} =
             :elixir_erl_clauses.match(&:elixir_erl_pass.translate_args/2, ex_args, scope)
@@ -544,8 +544,17 @@ defmodule Exception do
   defp format_application(module) do
     # We cannot use Application due to bootstrap issues
     case :application.get_application(module) do
-      {:ok, app} -> "(" <> Atom.to_string(app) <> ") "
-      :undefined -> ""
+      {:ok, app} ->
+        case :application.get_key(app, :vsn) do
+          {:ok, vsn} when is_list(vsn) ->
+            "(" <> Atom.to_string(app) <> " " <> List.to_string(vsn) <> ") "
+
+          _ ->
+            "(" <> Atom.to_string(app) <> ") "
+        end
+
+      :undefined ->
+        ""
     end
   end
 
@@ -1053,6 +1062,8 @@ end
 defmodule FunctionClauseError do
   defexception [:module, :function, :arity, :kind, :args, :clauses]
 
+  @clause_limit 10
+
   @impl true
   def message(exception) do
     case exception do
@@ -1098,37 +1109,46 @@ defmodule FunctionClauseError do
 
     mfa = Exception.format_mfa(module, function, arity)
 
-    formatted_args =
-      args
-      |> Enum.with_index(1)
-      |> Enum.map(fn {arg, i} ->
-        ["\n    # ", Integer.to_string(i), "\n    ", pad(inspect_fun.(arg)), "\n"]
-      end)
+    format_clause_fun = fn {args, guards} ->
+      code = Enum.reduce(guards, {function, [], args}, &{:when, [], [&2, &1]})
+      "    #{kind} " <> Macro.to_string(code, ast_fun) <> "\n"
+    end
 
-    formatted_clauses =
-      if clauses do
-        format_clause_fun = fn {args, guards} ->
-          code = Enum.reduce(guards, {function, [], args}, &{:when, [], [&2, &1]})
-          "    #{kind} " <> Macro.to_string(code, ast_fun) <> "\n"
-        end
-
-        top_10 =
-          clauses
-          |> Enum.take(10)
-          |> Enum.map(format_clause_fun)
-
-        [
-          "\nAttempted function clauses (showing #{length(top_10)} out of #{length(clauses)}):",
-          "\n\n",
-          top_10
-        ]
-      else
-        ""
-      end
-
-    "\n\nThe following arguments were given to #{mfa}:\n#{formatted_args}#{formatted_clauses}"
+    "\n\nThe following arguments were given to #{mfa}:\n" <>
+      "#{format_args(args, inspect_fun)}" <>
+      "#{format_clauses(clauses, format_clause_fun, @clause_limit)}"
   end
 
+  defp format_args(args, inspect_fun) do
+    args
+    |> Enum.with_index(1)
+    |> Enum.map(fn {arg, i} ->
+      [pad("\n# "), Integer.to_string(i), pad("\n"), pad(inspect_fun.(arg)), "\n"]
+    end)
+  end
+
+  defp format_clauses(clauses, format_clause_fun, limit)
+  defp format_clauses(nil, _, _), do: ""
+  defp format_clauses([], _, _), do: ""
+
+  defp format_clauses(clauses, format_clause_fun, limit) do
+    top_clauses =
+      clauses
+      |> Enum.take(limit)
+      |> Enum.map(format_clause_fun)
+
+    [
+      "\nAttempted function clauses (showing #{length(top_clauses)} out of #{length(clauses)}):",
+      "\n\n",
+      top_clauses,
+      non_visible_clauses(length(clauses) - limit)
+    ]
+  end
+
+  defp non_visible_clauses(n) when n <= 0, do: []
+  defp non_visible_clauses(1), do: ["    ...\n    (1 clause not shown)\n"]
+  defp non_visible_clauses(n), do: ["    ...\n    (#{n} clauses not shown)\n"]
+
   defp pad(string) do
     String.replace(string, "\n", "\n    ")
   end
