@@ -285,7 +285,7 @@ defmodule Kernel.Typespec do
       if is_atom(args) do
         []
       else
-        for(arg <- args, do: variable(arg))
+        :lists.map(&variable/1, args)
       end
 
     vars = :lists.filter(&match?({:var, _, _}, &1), args)
@@ -661,7 +661,7 @@ defmodule Kernel.Typespec do
         :elixir_errors.erl_warn(caller.line, caller.file, message)
 
         # This may be generating an invalid typespec but we need to generate it
-        # to avoid breaking existing code that was valid but only broke dialyzer
+        # to avoid breaking existing code that was valid but only broke Dialyzer
         {right, state} = typespec(expr, vars, caller, state)
         {{:ann_type, line(meta), [{:var, line(var_meta), var_name}, right]}, state}
 
@@ -680,7 +680,7 @@ defmodule Kernel.Typespec do
     :elixir_errors.erl_warn(caller.line, caller.file, message)
 
     # This may be generating an invalid typespec but we need to generate it
-    # to avoid breaking existing code that was valid but only broke dialyzer
+    # to avoid breaking existing code that was valid but only broke Dialyzer
     state = %{state | undefined_type_error_enabled?: false}
     {left, state} = typespec(left, vars, caller, state)
     state = %{state | undefined_type_error_enabled?: true}
@@ -725,14 +725,19 @@ defmodule Kernel.Typespec do
     # aliases in typespecs as compile time dependencies.
     remote = Macro.expand(remote, %{caller | function: {:typespec, 0}})
 
-    unless is_atom(remote) do
-      compile_error(caller, "invalid remote in typespec: #{Macro.to_string(orig)}")
-    end
+    cond do
+      not is_atom(remote) ->
+        compile_error(caller, "invalid remote in typespec: #{Macro.to_string(orig)}")
 
-    {remote_spec, state} = typespec(remote, vars, caller, state)
-    {name_spec, state} = typespec(name, vars, caller, state)
-    type = {remote_spec, meta, name_spec, args}
-    remote_type(type, vars, caller, state)
+      remote == caller.module ->
+        typespec({name, meta, args}, vars, caller, state)
+
+      true ->
+        {remote_spec, state} = typespec(remote, vars, caller, state)
+        {name_spec, state} = typespec(name, vars, caller, state)
+        type = {remote_spec, meta, name_spec, args}
+        remote_type(type, vars, caller, state)
+    end
   end
 
   # Handle tuples
@@ -828,7 +833,10 @@ defmodule Kernel.Typespec do
       false ->
         if state.undefined_type_error_enabled? and
              not Map.has_key?(state.defined_type_pairs, {name, arity}) do
-          compile_error(caller, "type #{name}/#{arity} undefined")
+          compile_error(
+            caller,
+            "type #{name}/#{arity} undefined (no such type in #{inspect(caller.module)})"
+          )
         end
 
         state =
