@@ -56,6 +56,9 @@ defmodule Module do
   If the behaviour changes or `URI.HTTP` does not implement
   one of the callbacks, a warning will be raised.
 
+  For detailed documentation, see the
+  [behaviour typespec documentation](typespecs.html#behaviours).
+
   ### `@impl`
 
   To aid in the correct implementation of behaviours, you may optionally declare
@@ -137,7 +140,7 @@ defmodule Module do
       end
 
   The Mix compiler automatically looks for calls to deprecated modules
-  and emit warnings during compilation, computed via `mix xref warnings`.
+  and emit warnings during compilation.
 
   Using the `@deprecated` attribute will also be reflected in the
   documentation of the given function and macro. You can choose between
@@ -346,7 +349,7 @@ defmodule Module do
 
   In addition to the built-in attributes outlined above, custom attributes may
   also be added. Custom attributes are expressed using the `@/1` operator followed
-  by a valid variable name. The value given to the custom attribute must be a valid 
+  by a valid variable name. The value given to the custom attribute must be a valid
   Elixir value:
 
       defmodule MyModule do
@@ -370,7 +373,7 @@ defmodule Module do
   When just a module is provided, the function is assumed to be
   `__after_compile__/2`.
 
-  Callbacks registered first will run last.
+  Callbacks will run in the order they are registered.
 
   #### Example
 
@@ -394,10 +397,11 @@ defmodule Module do
   When just a module is provided, the function/macro is assumed to be
   `__before_compile__/1`.
 
-  Callbacks registered first will run last. Any overridable definition
-  will be made concrete before the first callback runs. A definition may
-  be made overridable again in another before compile callback and it
-  will be made concrete one last time after after all callbacks run.
+  Callbacks will run in the order they are registered. Any overridable
+  definition will be made concrete before the first callback runs.
+  A definition may be made overridable again in another before compile
+  callback and it will be made concrete one last time after all callbacks
+  run.
 
   *Note*: unlike `@after_compile`, the callback function/macro must
   be placed in a separate module (because when the callback is invoked,
@@ -435,10 +439,6 @@ defmodule Module do
     * the list of quoted guards
     * the quoted function body
 
-  Note the hook receives the quoted arguments and it is invoked before
-  the function is stored in the module. So `Module.defines?/2` will return
-  `false` for the first clause of every function.
-
   If the function/macro being defined has multiple clauses, the hook will
   be called for each clause.
 
@@ -482,10 +482,10 @@ defmodule Module do
   below:
 
     * `@compile :debug_info` - includes `:debug_info` regardless of the
-      corresponding setting in `Code.compiler_options/1`
+      corresponding setting in `Code.get_compiler_option/1`
 
     * `@compile {:debug_info, false}` - disables `:debug_info` regardless
-      of the corresponding setting in `Code.compiler_options/1`
+      of the corresponding setting in `Code.get_compiler_option/1`
 
     * `@compile {:inline, some_fun: 2, other_fun: 3}` - inlines the given
       name/arity pairs. Inlining is applied locally, calls from another
@@ -495,6 +495,10 @@ defmodule Module do
       modules after compilation. Instead, the module will be loaded after
       it is dispatched to
 
+    * `@compile {:no_warn_undefined, Mod}` or
+      `@compile {:no_warn_undefined, {Mod, fun, arity}}` - does not warn if
+      the given module or the given `Mod.fun/arity` are not defined
+
   You can see a handful more options used by the Erlang compiler in
   the documentation for the [`:compile` module](http://www.erlang.org/doc/man/compile.html).
   '''
@@ -606,7 +610,7 @@ defmodule Module do
     assert_not_compiled!(__ENV__.function, module)
     :elixir_def.reset_last(module)
 
-    {value, binding, _env, _scope} =
+    {value, binding, _env} =
       :elixir.eval_quoted(quoted, binding, Keyword.put(opts, :module, module))
 
     {value, binding}
@@ -1066,7 +1070,7 @@ defmodule Module do
   """
   @spec make_overridable(module, [definition]) :: :ok
   def make_overridable(module, tuples) when is_atom(module) and is_list(tuples) do
-    assert_not_compiled!(__ENV__.function, module)
+    assert_not_readonly!(__ENV__.function, module)
 
     func = fn
       {function_name, arity} = tuple
@@ -1122,7 +1126,7 @@ defmodule Module do
     behaviour_definitions = bag_lookup_element(bag, {:accumulate, :behaviour}, 2)
 
     cond do
-      not Code.ensure_compiled?(behaviour) ->
+      Code.ensure_compiled(behaviour) != {:module, behaviour} ->
         {:error, "it was not defined"}
 
       not function_exported?(behaviour, :behaviour_info, 1) ->
@@ -1227,6 +1231,42 @@ defmodule Module do
     end
   end
 
+  @doc """
+  Checks if the given attribute has been defined.
+
+  An attribute is defined if it has been registered with `register_attribute/3`
+  or assigned a value. If an attribute has been deleted with `delete_attribute/2`
+  it is no longer considered defined.
+
+  This function can only be used on modules that have not yet been compiled.
+
+  ## Examples
+
+      defmodule MyModule do
+        @value 1
+        Module.register_attribute(__MODULE__, :other_value)
+        Module.put_attribute(__MODULE__, :another_value, 1)
+
+        Module.has_attribute?(__MODULE__, :value) #=> true
+        Module.has_attribute?(__MODULE__, :other_value) #=> true
+        Module.has_attribute?(__MODULE__, :another_value) #=> true
+
+        Module.has_attribute?(__MODULE__, :undefined) #=> false
+
+        Module.delete_attribute(__MODULE__, :value)
+        Module.has_attribute?(__MODULE__, :value) #=> false
+      end
+
+  """
+  @doc since: "1.10.0"
+  @spec has_attribute?(module, atom) :: boolean
+  def has_attribute?(module, key) when is_atom(module) and is_atom(key) do
+    assert_not_compiled!(__ENV__.function, module)
+    {set, _bag} = data_tables_for(module)
+
+    :ets.member(set, key)
+  end
+
   @doc """
   Deletes the module attribute that matches the given key.
 
@@ -1242,7 +1282,7 @@ defmodule Module do
   """
   @spec delete_attribute(module, atom) :: term
   def delete_attribute(module, key) when is_atom(module) and is_atom(key) do
-    assert_not_compiled!(__ENV__.function, module)
+    assert_not_readonly!(__ENV__.function, module)
     {set, bag} = data_tables_for(module)
 
     case :ets.lookup(set, key) do
@@ -1294,7 +1334,7 @@ defmodule Module do
   @spec register_attribute(module, atom, [{:accumulate, boolean}, {:persist, boolean}]) :: :ok
   def register_attribute(module, attribute, options)
       when is_atom(module) and is_atom(attribute) and is_list(options) do
-    assert_not_compiled!(__ENV__.function, module)
+    assert_not_readonly!(__ENV__.function, module)
     {set, bag} = data_tables_for(module)
 
     if Keyword.get(options, :persist) do
@@ -1304,6 +1344,9 @@ defmodule Module do
     if Keyword.get(options, :accumulate) do
       :ets.insert_new(set, {attribute, [], :accumulate}) ||
         :ets.update_element(set, attribute, {3, :accumulate})
+    else
+      :ets.insert_new(bag, {:warn_attributes, attribute})
+      :ets.insert_new(set, {attribute, nil, :unset})
     end
 
     :ok
@@ -1494,7 +1537,7 @@ defmodule Module do
     :ok
   end
 
-  defp check_behaviours(%{lexical_tracker: pid} = env, behaviours) do
+  defp check_behaviours(env, behaviours) do
     Enum.reduce(behaviours, %{}, fn behaviour, acc ->
       cond do
         not is_atom(behaviour) ->
@@ -1504,7 +1547,7 @@ defmodule Module do
           IO.warn(message, Macro.Env.stacktrace(env))
           acc
 
-        not Code.ensure_compiled?(behaviour) ->
+        Code.ensure_compiled(behaviour) != {:module, behaviour} ->
           message =
             "@behaviour #{inspect(behaviour)} does not exist (in module #{inspect(env.module)})"
 
@@ -1519,7 +1562,7 @@ defmodule Module do
           acc
 
         true ->
-          :elixir_lexical.record_remote(behaviour, nil, pid)
+          :elixir_env.trace({:require, [], behaviour, []}, env)
           optional_callbacks = behaviour_info(behaviour, :optional_callbacks)
           callbacks = behaviour_info(behaviour, :callbacks)
           Enum.reduce(callbacks, acc, &add_callback(&1, behaviour, env, optional_callbacks, &2))
@@ -1771,11 +1814,11 @@ defmodule Module do
       [{_, _, :accumulate}] ->
         :lists.reverse(bag_lookup_element(bag, {:accumulate, key}, 2))
 
-      [{_, val, nil}] ->
+      [{_, val, line}] when is_integer(line) ->
+        :ets.update_element(set, key, {3, :used})
         val
 
       [{_, val, _}] ->
-        :ets.update_element(set, key, {3, nil})
         val
 
       [] when is_integer(line) ->
@@ -1796,7 +1839,7 @@ defmodule Module do
   # Used internally by Kernel's @.
   # This function is private and must be used only internally.
   def __put_attribute__(module, key, value, line) when is_atom(key) do
-    assert_not_compiled!(__ENV__.function, module)
+    assert_not_readonly!(__ENV__.function, module)
     {set, bag} = data_tables_for(module)
     value = preprocess_attribute(key, value)
     put_attribute(module, key, value, line, set, bag)
@@ -1846,7 +1889,7 @@ defmodule Module do
     catch
       :error, :badarg ->
         :ets.insert(set, {:on_load, value, line})
-        :ets.insert(bag, {:attributes, :on_load})
+        :ets.insert(bag, {:warn_attributes, :on_load})
     else
       _ -> raise ArgumentError, "the @on_load attribute can only be set once per module"
     end
@@ -1858,7 +1901,7 @@ defmodule Module do
     catch
       :error, :badarg ->
         :ets.insert(set, {key, value, line})
-        :ets.insert(bag, {:attributes, key})
+        :ets.insert(bag, {:warn_attributes, key})
     else
       :accumulate -> :ets.insert(bag, {{:accumulate, key}, value})
       _ -> :ets.insert(set, {key, value, line})
@@ -2041,6 +2084,22 @@ defmodule Module do
             assert_not_compiled_message(function_name_arity, module, extra_msg)
   end
 
+  defp assert_not_readonly!({function_name, arity}, module) do
+    case :elixir_module.mode(module) do
+      :all ->
+        :ok
+
+      :readonly ->
+        raise ArgumentError,
+              "could not call Module.#{function_name}/#{arity} because the module " <>
+                "#{inspect(module)} is in read-only mode (@after_compile)"
+
+      :closed ->
+        raise ArgumentError,
+              assert_not_compiled_message({function_name, arity}, module, "")
+    end
+  end
+
   defp assert_not_compiled_message({function_name, arity}, module, extra_msg) do
     mfa = "Module.#{function_name}/#{arity}"
 
