@@ -1,7 +1,6 @@
 %% Translate Elixir quoted expressions to Erlang Abstract Format.
 -module(elixir_erl_pass).
--export([translate/2, translate_arg/3, translate_args/2]).
--import(elixir_erl_var, [mergev/2, mergec/2]).
+-export([translate/2, translate_args/2]).
 -include("elixir.hrl").
 
 %% =
@@ -16,7 +15,7 @@ translate({'=', Meta, [Left, Right]}, S) ->
     {TLeft, #elixir_erl{extra_guards=ExtraGuards, context=Context} = SL0}
         when ExtraGuards =/= [], Context =/= match ->
       SL1 = SL0#elixir_erl{extra_guards=[]},
-      {ResultVarName, _, SL2} = elixir_erl_var:build('_', SL1),
+      {ResultVarName, SL2} = elixir_erl_var:build('_', SL1),
       Match = {match, ?ann(Meta), TLeft, TRight},
       Generated = ?ann(?generated(Meta)),
       ResultVar = {var, Generated, ResultVarName},
@@ -66,10 +65,10 @@ translate({'__block__', Meta, Args}, S) when is_list(Args) ->
 translate({'__CALLER__', Meta, Atom}, S) when is_atom(Atom) ->
   {{var, ?ann(Meta), '__CALLER__'}, S#elixir_erl{caller=true}};
 
-translate({'__STACKTRACE__', Meta, Atom}, S = #elixir_erl{stacktrace={Var, _}}) when is_atom(Atom) ->
-  {{var, ?ann(Meta), Var}, S#elixir_erl{stacktrace={Var, true}}};
+translate({'__STACKTRACE__', Meta, Atom}, S = #elixir_erl{stacktrace=Var}) when is_atom(Atom) ->
+  {{var, ?ann(Meta), Var}, S};
 
-translate({'super', Meta, Args}, S) ->
+translate({'super', Meta, Args}, S) when is_list(Args) ->
   %% In the expanded AST, super is used to invoke a function
   %% in the current module originated from a default clause
   %% or a super call.
@@ -96,14 +95,22 @@ translate({'&', Meta, [{'/', _, [{{'.', _, [Remote, Fun]}, _, []}, Arity]}]}, S)
   {{'fun', Ann, {function, TRemote, TFun, TArity}}, SR};
 translate({'&', Meta, [{'/', _, [{Fun, _, Atom}, Arity]}]}, S)
     when is_atom(Fun), is_atom(Atom), is_integer(Arity) ->
-  {{'fun', ?ann(Meta), {function, Fun, Arity}}, S};
+  case S of
+    #elixir_erl{expand_captures=true} ->
+      {Vars, SV} = lists:mapfoldl(fun(_, Acc) ->
+        {Var, _, AccS} = elixir_erl_var:assign(Meta, Acc),
+        {Var, AccS}
+      end, S, tl(lists:seq(0, Arity))),
+      translate({'fn', Meta, [{'->', Meta, [Vars, {Fun, Meta, Vars}]}]}, SV);
+
+    #elixir_erl{expand_captures=false} ->
+      {{'fun', ?ann(Meta), {function, Fun, Arity}}, S}
+  end;
 
 translate({fn, Meta, Clauses}, S) ->
   Transformer = fun({'->', CMeta, [ArgsWithGuards, Expr]}, Acc) ->
     {Args, Guards} = elixir_utils:extract_splat_guards(ArgsWithGuards),
-    {TClause, TS} = elixir_erl_clauses:clause(CMeta, fun translate_fn_match/2,
-                                              Args, Expr, Guards, Acc),
-    {TClause, elixir_erl_var:mergec(S, TS)}
+    elixir_erl_clauses:clause(CMeta, fun translate_fn_match/2, Args, Expr, Guards, Acc)
   end,
   {TClauses, NS} = lists:mapfoldl(Transformer, S, Clauses),
   {{'fun', ?ann(Meta), {clauses, TClauses}}, NS};
@@ -119,9 +126,9 @@ translate({'cond', CondMeta, [[{do, Clauses}]]}, S) ->
       true -> Meta
     end,
 
-  Error = {{'.', Meta, [erlang, error]}, [], [cond_clause]},
-  Case = build_cond_clauses([H | T], Error, FirstMeta),
-  translate(replace_case_meta(CondMeta, Case), S);
+  Error = {{'.', Meta, [erlang, error]}, Meta, [cond_clause]},
+  {Case, SC} = build_cond_clauses([H | T], Error, FirstMeta, S),
+  translate(replace_case_meta(CondMeta, Case), SC);
 
 %% Case
 
@@ -135,19 +142,19 @@ translate({'try', Meta, [Opts]}, S) ->
   {TDo, SB} = translate(Do, S),
 
   Catch = [Tuple || {X, _} = Tuple <- Opts, X == 'rescue' orelse X == 'catch'],
-  {TCatch, SC} = elixir_erl_try:clauses(Meta, Catch, mergec(S, SB)),
+  {TCatch, SC} = elixir_erl_try:clauses(Meta, Catch, SB),
 
   {TAfter, SA} = case lists:keyfind('after', 1, Opts) of
     {'after', After} ->
-      {TBlock, SAExtracted} = translate(After, mergec(S, SC)),
+      {TBlock, SAExtracted} = translate(After, SC),
       {unblock(TBlock), SAExtracted};
     false ->
-      {[], mergec(S, SC)}
+      {[], SC}
   end,
 
   Else = elixir_erl_clauses:get_clauses(else, Opts, match),
-  {TElse, SE} = elixir_erl_clauses:clauses(Else, mergec(S, SA)),
-  {{'try', ?ann(Meta), unblock(TDo), TElse, TCatch, TAfter}, mergec(S, SE)};
+  {TElse, SE} = elixir_erl_clauses:clauses(Else, SA),
+  {{'try', ?ann(Meta), unblock(TDo), TElse, TCatch, TAfter}, SE};
 
 %% Receive
 
@@ -174,32 +181,26 @@ translate({for, Meta, [_ | _] = Args}, S) ->
 translate({with, Meta, [_ | _] = Args}, S) ->
   {Exprs, [{do, Do} | Opts]} = elixir_utils:split_last(Args),
   {ElseClause, SE} = translate_with_else(Meta, Opts, S),
-  {With, SD} = translate_with_do(Exprs, Do, ElseClause, elixir_erl_var:mergec(S, SE)),
-  {With, elixir_erl_var:mergec(S, SD)};
+  translate_with_do(Exprs, Do, ElseClause, SE);
 
 %% Variables
 
-translate({'^', Meta, [{Name, VarMeta, Kind}]}, #elixir_erl{context=match} = S) when is_atom(Name), is_atom(Kind) ->
-  Tuple = {Name, elixir_utils:var_context(VarMeta, Kind)},
-  {ok, {_Counter, Value}} = maps:find(Tuple, S#elixir_erl.backup_vars),
-
-  PAnn = ?ann(?generated(Meta)),
-  PVar = {var, PAnn, Value},
+translate({'^', _, [{Name, VarMeta, Kind}]}, #elixir_erl{context=match} = S) when is_atom(Name), is_atom(Kind) ->
+  {Var, VS} = elixir_erl_var:translate(VarMeta, Name, Kind, S),
 
   case S#elixir_erl.extra of
     pin_guard ->
-      {TVar, TS} = elixir_erl_var:translate(VarMeta, Name, elixir_utils:var_context(VarMeta, Kind), S),
-      Guard = {op, PAnn, '=:=', PVar, TVar},
-      {TVar, TS#elixir_erl{extra_guards=[Guard | TS#elixir_erl.extra_guards]}};
+      {PinVarName, PS} = elixir_erl_var:build('_', VS),
+      Ann = ?ann(?generated(VarMeta)),
+      PinVar = {var, Ann, PinVarName},
+      Guard = {op, Ann, '=:=', Var, PinVar},
+      {PinVar, PS#elixir_erl{extra_guards=[Guard | PS#elixir_erl.extra_guards]}};
     _ ->
-      {PVar, S}
+      {Var, VS}
   end;
 
-translate({'_', Meta, Kind}, #elixir_erl{context=match} = S) when is_atom(Kind) ->
-  {{var, ?ann(Meta), '_'}, S};
-
 translate({Name, Meta, Kind}, S) when is_atom(Name), is_atom(Kind) ->
-  elixir_erl_var:translate(Meta, Name, elixir_utils:var_context(Meta, Kind), S);
+  elixir_erl_var:translate(Meta, Name, Kind, S);
 
 %% Local calls
 
@@ -213,7 +214,7 @@ translate({Name, Meta, Args}, S) when is_atom(Name), is_list(Meta), is_list(Args
 translate({{'.', _, [Left, Right]}, Meta, []}, S)
     when is_tuple(Left), is_atom(Right), is_list(Meta) ->
   {TLeft, SL}  = translate(Left, S),
-  {Var, _, SV} = elixir_erl_var:build('_', SL),
+  {Var, SV} = elixir_erl_var:build('_', SL),
 
   Ann = ?ann(Meta),
   Generated = erl_anno:set_generated(true, Ann),
@@ -244,17 +245,13 @@ translate({{'.', _, [Left, Right]}, Meta, Args}, S)
 
 translate({{'.', _, [Expr]}, Meta, Args}, S) when is_list(Args) ->
   {TExpr, SE} = translate(Expr, S),
-  {TArgs, SA} = translate_args(Args, mergec(S, SE)),
-  {{call, ?ann(Meta), TExpr, TArgs}, mergev(SE, SA)};
+  {TArgs, SA} = translate_args(Args, SE),
+  {{call, ?ann(Meta), TExpr, TArgs}, SA};
 
 %% Literals
 
 translate(List, S) when is_list(List) ->
-  Fun = case S#elixir_erl.context of
-    match -> fun translate/2;
-    _     -> fun(X, Acc) -> translate_arg(X, Acc, S) end
-  end,
-  translate_list(List, Fun, S, []);
+  translate_list(List, S, []);
 
 translate({Left, Right}, S) ->
   {TArgs, SE} = translate_args([Left, Right], S),
@@ -271,13 +268,13 @@ translate_case(Meta, Expr, Opts, S) ->
   {TClauses, SC} = elixir_erl_clauses:clauses(Clauses, SE),
   {{'case', ?ann(Meta), TExpr, TClauses}, SC}.
 
-translate_list([{'|', _, [_, _]=Args}], Fun, Acc, List) ->
-  {[TLeft, TRight], TAcc} = lists:mapfoldl(Fun, Acc, Args),
+translate_list([{'|', _, [_, _]=Args}], Acc, List) ->
+  {[TLeft, TRight], TAcc} = lists:mapfoldl(fun translate/2, Acc, Args),
   {build_list([TLeft | List], TRight), TAcc};
-translate_list([H | T], Fun, Acc, List) ->
-  {TH, TAcc} = Fun(H, Acc),
-  translate_list(T, Fun, TAcc, [TH | List]);
-translate_list([], _Fun, Acc, List) ->
+translate_list([H | T], Acc, List) ->
+  {TH, TAcc} = translate(H, Acc),
+  translate_list(T, TAcc, [TH | List]);
+translate_list([], Acc, List) ->
   {build_list(List, {nil, 0}), Acc}.
 
 build_list([H | T], Acc) ->
@@ -295,18 +292,8 @@ translate_fn_match(Arg, S) ->
 
 %% Translate args
 
-translate_arg(Arg, Acc, S) when is_number(Arg); is_atom(Arg); is_binary(Arg); is_pid(Arg); is_function(Arg) ->
-  {TArg, _} = translate(Arg, S),
-  {TArg, Acc};
-translate_arg(Arg, Acc, S) ->
-  {TArg, TAcc} = translate(Arg, mergec(S, Acc)),
-  {TArg, mergev(Acc, TAcc)}.
-
-translate_args(Args, #elixir_erl{context=match} = S) ->
-  lists:mapfoldl(fun translate/2, S, Args);
-
 translate_args(Args, S) ->
-  lists:mapfoldl(fun(X, Acc) -> translate_arg(X, Acc, S) end, S, Args).
+  lists:mapfoldl(fun translate/2, S, Args).
 
 %% Translate blocks
 
@@ -329,32 +316,32 @@ translate_block([H | T], Acc, S) ->
 
 %% Cond
 
-build_cond_clauses([{'->', NewMeta, [[Condition], Body]} | T], Acc, OldMeta) ->
-  {NewCondition, Truthy, Other} = build_truthy_clause(NewMeta, Condition, Body),
+build_cond_clauses([{'->', NewMeta, [[Condition], Body]} | T], Acc, OldMeta, S) ->
+  {NewCondition, Truthy, Other, ST} = build_truthy_clause(NewMeta, Condition, Body, S),
   Falsy = {'->', OldMeta, [[Other], Acc]},
   Case = {'case', NewMeta, [NewCondition, [{do, [Truthy, Falsy]}]]},
-  build_cond_clauses(T, Case, NewMeta);
-build_cond_clauses([], Acc, _) ->
-  Acc.
+  build_cond_clauses(T, Case, NewMeta, ST);
+build_cond_clauses([], Acc, _, S) ->
+  {Acc, S}.
 
 replace_case_meta(Meta, {'case', _, Args}) ->
   {'case', Meta, Args};
 replace_case_meta(_Meta, Other) ->
   Other.
 
-build_truthy_clause(Meta, Condition, Body) ->
+build_truthy_clause(Meta, Condition, Body, S) ->
   case returns_boolean(Condition, Body) of
     {NewCondition, NewBody} ->
-      {NewCondition, {'->', Meta, [[true], NewBody]}, false};
+      {NewCondition, {'->', Meta, [[true], NewBody]}, false, S};
     false ->
-      Var = {'cond', [], ?var_context},
+      {Var, _, SV} = elixir_erl_var:assign(Meta, S),
       Head = {'when', [], [Var,
         {{'.', [], [erlang, 'andalso']}, [], [
           {{'.', [], [erlang, '/=']}, [], [Var, nil]},
           {{'.', [], [erlang, '/=']}, [], [Var, false]}
         ]}
       ]},
-      {Condition, {'->', Meta, [[Head], Body]}, {'_', [], nil}}
+      {Condition, {'->', Meta, [[Head], Body]}, {'_', [], nil}, SV}
   end.
 
 %% In case a variable is defined to match in a condition
@@ -378,24 +365,21 @@ returns_boolean(Condition, Body) ->
 
 translate_with_else(Meta, [], S) ->
   Ann = ?ann(?generated(Meta)),
-  {VarName, _, SC} = elixir_erl_var:build('_', S),
+  {VarName, SC} = elixir_erl_var:build('_', S),
   Var = {var, Ann, VarName},
   {{clause, Ann, [Var], [], [Var]}, SC};
 translate_with_else(Meta, [{else, [{'->', _, [[{Var, VarMeta, Kind}], Clause]}]}], S) when is_atom(Var), is_atom(Kind) ->
   Ann = ?ann(?generated(Meta)),
-  Ctx = elixir_utils:var_context(VarMeta, Kind),
-  {ElseVarErl, SV} = elixir_erl_var:assign(Meta, Var, Ctx, S),
-  {TranslatedClause, SC} = elixir_erl_pass:translate(Clause, SV),
+  {ElseVarErl, SV} = elixir_erl_var:translate(VarMeta, Var, Kind, S#elixir_erl{context=match}),
+  {TranslatedClause, SC} = elixir_erl_pass:translate(Clause, SV#elixir_erl{context=nil}),
   {{clause, Ann, [ElseVarErl], [], [TranslatedClause]}, SC};
 translate_with_else(Meta, [{else, Else}], S) ->
   Generated = ?generated(Meta),
-  ElseVarEx = {else, Generated, ?var_context},
-  {ElseVarErl, SV} = elixir_erl_var:assign(Generated, else, ?var_context, S),
+  {ElseVarEx, ElseVarErl, SE} = elixir_erl_var:assign(Generated, S),
+  {RaiseVar, _, SV} = elixir_erl_var:assign(Generated, SE),
 
-  RaiseVar = {catch_all, Generated, ?var_context},
   RaiseExpr = {{'.', Generated, [erlang, error]}, Generated, [{with_clause, RaiseVar}]},
   RaiseClause = {'->', Generated, [[RaiseVar], RaiseExpr]},
-
   GeneratedElse = [build_generated_clause(Generated, ElseClause) || ElseClause <- Else],
 
   Case = {'case', Generated, [ElseVarEx, [{do, GeneratedElse ++ [RaiseClause]}]]},
@@ -425,8 +409,8 @@ translate_with_do([{'<-', Meta, [Left, Expr]} | Rest], Do, Else, S) ->
   {Args, Guards} = elixir_utils:extract_guards(Left),
   {TExpr, SR} = elixir_erl_pass:translate(Expr, S),
   {TArgs, SA} = elixir_erl_clauses:match(fun elixir_erl_pass:translate/2, Args, SR),
-  TGuards = elixir_erl_clauses:guards(Guards, [], SA),
-  {TBody, SB} = translate_with_do(Rest, Do, Else, SA),
+  TGuards = elixir_erl_clauses:guards(Guards, SA#elixir_erl.extra_guards, SA),
+  {TBody, SB} = translate_with_do(Rest, Do, Else, SA#elixir_erl{extra_guards=[]}),
 
   Clause = {clause, ?ann(Meta), [TArgs], TGuards, unblock(TBody)},
   {{'case', ?ann(?generated(Meta)), TExpr, [Clause, Else]}, SB};
@@ -439,12 +423,6 @@ translate_with_do([], Do, _Else, S) ->
 
 %% Maps and structs
 
-translate_map(Meta, [{'|', _Meta, [Update, Assocs]}], S) ->
-  {TUpdate, US} = translate_arg(Update, S, S),
-  translate_map(Meta, Assocs, {ok, TUpdate}, US);
-translate_map(Meta, Assocs, S) ->
-  translate_map(Meta, Assocs, none, S).
-
 translate_struct_var_name(Meta, Name, Args, S0) ->
   {{map, MapAnn, TArgs0}, S1} = translate_struct(Meta, Name, Args, S0),
   {TArgs1, S2} = generate_struct_name_guard(TArgs0, [], S1),
@@ -453,7 +431,7 @@ translate_struct_var_name(Meta, Name, Args, S0) ->
 translate_struct(Meta, Name, {'%{}', _, [{'|', _, [Update, Assocs]}]}, S) ->
   Ann = ?ann(Meta),
   Generated = erl_anno:set_generated(true, Ann),
-  {VarName, _, VS} = elixir_erl_var:build('_', S),
+  {VarName, VS} = elixir_erl_var:build('_', S),
 
   Var = {var, Ann, VarName},
   Map = {map, Ann, [{map_field_exact, Ann, {atom, Ann, '__struct__'}, {atom, Ann, Name}}]},
@@ -461,8 +439,8 @@ translate_struct(Meta, Name, {'%{}', _, [{'|', _, [Update, Assocs]}]}, S) ->
   Match = {match, Ann, Var, Map},
   Error = {tuple, Ann, [{atom, Ann, badstruct}, {atom, Ann, Name}, Var]},
 
-  {TUpdate, US} = translate_arg(Update, VS, VS),
-  {TAssocs, TS} = translate_map(Meta, Assocs, {ok, Var}, US),
+  {TUpdate, TU} = translate(Update, VS),
+  {TAssocs, TS} = translate_map(Meta, Assocs, {ok, Var}, TU),
 
   {{'case', Generated, TUpdate, [
     {clause, Ann, [Match], [], [TAssocs]},
@@ -471,32 +449,28 @@ translate_struct(Meta, Name, {'%{}', _, [{'|', _, [Update, Assocs]}]}, S) ->
 translate_struct(Meta, Name, {'%{}', _, Assocs}, S) ->
   translate_map(Meta, Assocs ++ [{'__struct__', Name}], none, S).
 
+translate_map(Meta, [{'|', _Meta, [Update, Assocs]}], S) ->
+  {TUpdate, SU} = translate(Update, S),
+  translate_map(Meta, Assocs, {ok, TUpdate}, SU);
+translate_map(Meta, Assocs, S) ->
+  translate_map(Meta, Assocs, none, S).
+
 translate_map(Meta, Assocs, TUpdate, #elixir_erl{extra=Extra} = S) ->
-  {Op, KeyFun, ValFun} = translate_key_val_op(TUpdate, S),
+  Op = translate_key_val_op(TUpdate, S),
   Ann = ?ann(Meta),
 
   {TArgs, SA} = lists:mapfoldl(fun({Key, Value}, Acc0) ->
-    {TKey, Acc1} = KeyFun(Key, Acc0),
-    {TValue, Acc2} = ValFun(Value, Acc1#elixir_erl{extra=Extra}),
+    {TKey, Acc1} = translate(Key, Acc0#elixir_erl{extra=map_key}),
+    {TValue, Acc2} = translate(Value, Acc1#elixir_erl{extra=Extra}),
     {{Op, ?ann(Meta), TKey, TValue}, Acc2}
   end, S, Assocs),
 
   build_map(Ann, TUpdate, TArgs, SA).
 
-translate_key_val_op(_TUpdate, #elixir_erl{extra=map_key}) ->
-  {map_field_assoc,
-    fun(X, Acc) -> translate(X, Acc#elixir_erl{extra=map_key}) end,
-    fun translate/2};
-translate_key_val_op(_TUpdate, #elixir_erl{context=match}) ->
-  {map_field_exact,
-    fun(X, Acc) -> translate(X, Acc#elixir_erl{extra=map_key}) end,
-    fun translate/2};
-translate_key_val_op(TUpdate, S) ->
-  KS = S#elixir_erl{extra=map_key},
-  Op = if TUpdate == none -> map_field_assoc; true -> map_field_exact end,
-  {Op,
-    fun(X, Acc) -> translate_arg(X, Acc, KS) end,
-    fun(X, Acc) -> translate_arg(X, Acc, S) end}.
+translate_key_val_op(_TUpdate, #elixir_erl{extra=map_key}) -> map_field_assoc;
+translate_key_val_op(_TUpdate, #elixir_erl{context=match}) -> map_field_exact;
+translate_key_val_op(none, _) -> map_field_assoc;
+translate_key_val_op(_, _) -> map_field_exact.
 
 build_map(Ann, {ok, TUpdate}, TArgs, SA) -> {{map, Ann, TUpdate, TArgs}, SA};
 build_map(Ann, none, TArgs, SA) -> {{map, Ann, TArgs}, SA}.
@@ -504,18 +478,15 @@ build_map(Ann, none, TArgs, SA) -> {{map, Ann, TArgs}, SA}.
 %% Translate bitstrings
 
 translate_bitstring(Meta, Args, S) ->
-  case S#elixir_erl.context of
-    match -> build_bitstr(fun translate/2, Args, Meta, S, []);
-    _ -> build_bitstr(fun(X, Acc) -> translate_arg(X, Acc, S) end, Args, Meta, S, [])
-  end.
+  build_bitstr(Args, Meta, S, []).
 
-build_bitstr(Fun, [{'::', _, [H, V]} | T], Meta, S, Acc) ->
+build_bitstr([{'::', _, [H, V]} | T], Meta, S, Acc) ->
   {Size, Types} = extract_bit_info(V, S#elixir_erl{context=nil}),
-  build_bitstr(Fun, T, Meta, S, Acc, H, Size, Types);
-build_bitstr(_Fun, [], Meta, S, Acc) ->
+  build_bitstr(T, Meta, S, Acc, H, Size, Types);
+build_bitstr([], Meta, S, Acc) ->
   {{bin, ?ann(Meta), lists:reverse(Acc)}, S}.
 
-build_bitstr(Fun, T, Meta, S, Acc, H, default, Types) when is_binary(H) ->
+build_bitstr(T, Meta, S, Acc, H, default, Types) when is_binary(H) ->
   Element =
     case requires_utf_conversion(Types) of
       false ->
@@ -526,11 +497,11 @@ build_bitstr(Fun, T, Meta, S, Acc, H, default, Types) when is_binary(H) ->
         %% UTF types require conversion.
         {bin_element, ?ann(Meta), {string, 0, elixir_utils:characters_to_list(H)}, default, Types}
     end,
-  build_bitstr(Fun, T, Meta, S, [Element | Acc]);
+  build_bitstr(T, Meta, S, [Element | Acc]);
 
-build_bitstr(Fun, T, Meta, S, Acc, H, Size, Types) ->
-  {Expr, NS} = Fun(H, S),
-  build_bitstr(Fun, T, Meta, NS, [{bin_element, ?ann(Meta), Expr, Size, Types} | Acc]).
+build_bitstr(T, Meta, S, Acc, H, Size, Types) ->
+  {Expr, NS} = translate(H, S),
+  build_bitstr(T, Meta, NS, [{bin_element, ?ann(Meta), Expr, Size, Types} | Acc]).
 
 requires_utf_conversion([bitstring | _]) -> false;
 requires_utf_conversion([binary | _]) -> false;
@@ -557,97 +528,71 @@ extract_bit_type({Other, _, []}, Acc) ->
 %% Optimizations that are specific to Erlang and change
 %% the format of the AST.
 
-translate_remote('Elixir.Access' = Mod, get, Meta, [Container, Value], S) ->
-  Ann = ?ann(Meta),
-  {TArgs, SA} = translate_args([Container, Value, nil], S),
-  {?remote(Ann, Mod, get, TArgs), SA};
 translate_remote('Elixir.String.Chars', to_string, Meta, [Arg], S) ->
-  case is_always_string(Arg) of
-    true ->
-      translate(Arg, S);
-    false ->
-      {TArg, TS} = translate(Arg, S),
-      {VarName, _, VS} = elixir_erl_var:build(rewrite, TS),
+  {TArg, TS} = translate(Arg, S),
+  {VarName, VS} = elixir_erl_var:build('_', TS),
 
-      Generated = ?ann(?generated(Meta)),
-      Var = {var, Generated, VarName},
-      Guard = ?remote(Generated, erlang, is_binary, [Var]),
-      Slow = ?remote(Generated, 'Elixir.String.Chars', to_string, [Var]),
-      Fast = Var,
+  Generated = ?ann(?generated(Meta)),
+  Var = {var, Generated, VarName},
+  Guard = ?remote(Generated, erlang, is_binary, [Var]),
+  Slow = ?remote(Generated, 'Elixir.String.Chars', to_string, [Var]),
+  Fast = Var,
 
-      {{'case', Generated, TArg, [
-        {clause, Generated, [Var], [[Guard]], [Fast]},
-        {clause, Generated, [Var], [], [Slow]}
-      ]}, VS}
-  end;
+  {{'case', Generated, TArg, [
+    {clause, Generated, [Var], [[Guard]], [Fast]},
+    {clause, Generated, [Var], [], [Slow]}
+  ]}, VS};
 translate_remote(maps, put, Meta, [Key, Value, Map], S) ->
   Ann = ?ann(Meta),
-  {TExpr, ES} =
-    case translate_args([Key, Value, Map], S) of
-      {[TKey, TValue, {map, _, InnerMap, Pairs}], TS} ->
-        {{map, Ann, InnerMap, Pairs ++ [{map_field_assoc, Ann, TKey, TValue}]}, TS};
 
-      {[TKey, TValue, {map, _, Pairs}], TS} ->
-        {{map, Ann, Pairs ++ [{map_field_assoc, Ann, TKey, TValue}]}, TS};
+  case translate_args([Key, Value, Map], S) of
+    {[TKey, TValue, {map, _, InnerMap, Pairs}], TS} ->
+      {{map, Ann, InnerMap, Pairs ++ [{map_field_assoc, Ann, TKey, TValue}]}, TS};
 
-      {[TKey, TValue, TMap], TS} ->
-        {{map, Ann, TMap, [{map_field_assoc, Ann, TKey, TValue}]}, TS}
-    end,
-  {TExpr, mergev(S, ES)};
+    {[TKey, TValue, {map, _, Pairs}], TS} ->
+      {{map, Ann, Pairs ++ [{map_field_assoc, Ann, TKey, TValue}]}, TS};
+
+    {[TKey, TValue, TMap], TS} ->
+      {{map, Ann, TMap, [{map_field_assoc, Ann, TKey, TValue}]}, TS}
+  end;
 translate_remote(maps, merge, Meta, [Map1, Map2], S) ->
   Ann = ?ann(Meta),
-  {TExpr, ES} =
-    case translate_args([Map1, Map2], S) of
-      {[{map, _, Pairs1}, {map, _, Pairs2}], TS} ->
-        {{map, Ann, Pairs1 ++ Pairs2}, TS};
 
-      {[{map, _, InnerMap1, Pairs1}, {map, _, Pairs2}], TS} ->
-        {{map, Ann, InnerMap1, Pairs1 ++ Pairs2}, TS};
+  case translate_args([Map1, Map2], S) of
+    {[{map, _, Pairs1}, {map, _, Pairs2}], TS} ->
+      {{map, Ann, Pairs1 ++ Pairs2}, TS};
 
-      {[TMap1, {map, _, Pairs2}], TS} ->
-        {{map, Ann, TMap1, Pairs2}, TS};
+    {[{map, _, InnerMap1, Pairs1}, {map, _, Pairs2}], TS} ->
+      {{map, Ann, InnerMap1, Pairs1 ++ Pairs2}, TS};
 
-      {[TMap1, TMap2], TS} ->
-        {{call, Ann, {remote, Ann, {atom, Ann, maps}, {atom, Ann, merge}}, [TMap1, TMap2]}, TS}
-    end,
-  {TExpr, mergev(S, ES)};
+    {[TMap1, {map, _, Pairs2}], TS} ->
+      {{map, Ann, TMap1, Pairs2}, TS};
+
+    {[TMap1, TMap2], TS} ->
+      {{call, Ann, {remote, Ann, {atom, Ann, maps}, {atom, Ann, merge}}, [TMap1, TMap2]}, TS}
+  end;
 translate_remote(Left, Right, Meta, Args, S) ->
   {TLeft, SL} = translate(Left, S),
-  {TArgs, SA} = translate_args(Args, mergec(S, SL)),
+  {TArgs, SA} = lists:mapfoldl(fun translate/2, SL, Args),
 
   Ann    = ?ann(Meta),
   Arity  = length(Args),
   TRight = {atom, Ann, Right},
-  SC = mergev(SL, SA),
 
   %% Rewrite Erlang function calls as operators so they
   %% work in guards, matches and so on.
   case (Left == erlang) andalso elixir_utils:guard_op(Right, Arity) of
     true ->
       case TArgs of
-        [TOne]       -> {{op, Ann, Right, TOne}, SC};
-        [TOne, TTwo] -> {{op, Ann, Right, TOne, TTwo}, SC}
+        [TOne]       -> {{op, Ann, Right, TOne}, SA};
+        [TOne, TTwo] -> {{op, Ann, Right, TOne, TTwo}, SA}
       end;
     false ->
-      {{call, Ann, {remote, Ann, TLeft, TRight}, TArgs}, SC}
+      {{call, Ann, {remote, Ann, TLeft, TRight}, TArgs}, SA}
   end.
 
-is_always_string({{'.', _, [Module, Function]}, _, Args}) ->
-  is_always_string(Module, Function, length(Args));
-%% Binary literals were already excluded in earlier passes.
-is_always_string(_Ast) ->
-  false.
-
-is_always_string('Elixir.Enum', join, _) -> true;
-is_always_string('Elixir.Enum', map_join, _) -> true;
-is_always_string('Elixir.Kernel', inspect, _) -> true;
-is_always_string('Elixir.Macro', to_string, _) -> true;
-is_always_string('Elixir.String.Chars', to_string, _) -> true;
-is_always_string('Elixir.Path', join, _) -> true;
-is_always_string(_Module, _Function, _Args) -> false.
-
 generate_struct_name_guard([{map_field_exact, Ann, {atom, _, '__struct__'} = Key, Var} | Rest], Acc, S0) ->
-  {ModuleVarName, _, S1} = elixir_erl_var:build('_', S0),
+  {ModuleVarName, S1} = elixir_erl_var:build('_', S0),
   Generated = erl_anno:set_generated(true, Ann),
   ModuleVar = {var, Generated, ModuleVarName},
   Match = {match, Generated, ModuleVar, Var},
