@@ -529,8 +529,8 @@ defmodule Code.Formatter.CallsTest do
       end
       """
 
-      # Doesn't preserve this because only the beginning has a newline
-      assert_format "call(\nfoo, bar, baz)", "call(foo, bar, baz)"
+      # Doesn't preserve this because only the ending has a newline
+      assert_format "call(foo, bar, baz\n)", "call(foo, bar, baz)"
 
       # Doesn't preserve because there are no args
       bad = """
@@ -554,7 +554,6 @@ defmodule Code.Formatter.CallsTest do
       )
       """
 
-      # Doesn't preserve this because only the beginning has a newline
       assert_format bad, """
       call(%{
         key: :value
@@ -728,8 +727,8 @@ defmodule Code.Formatter.CallsTest do
       )
       """
 
-      # Doesn't preserve this because only the beginning has a newline
-      assert_format "Remote.call(\nfoo, bar, baz)", "Remote.call(foo, bar, baz)"
+      # Doesn't preserve this because only the ending has a newline
+      assert_format "Remote.call(foo, bar, baz\n)", "Remote.call(foo, bar, baz)"
 
       assert_same """
       Remote.call(
@@ -837,8 +836,8 @@ defmodule Code.Formatter.CallsTest do
       )
       """
 
-      # Doesn't preserve this because only the beginning has a newline
-      assert_format "call.(\nfoo, bar, baz)", "call.(foo, bar, baz)"
+      # Doesn't preserve this because only the ending has a newline
+      assert_format "call.(foo, bar, baz\n)", "call.(foo, bar, baz)"
     end
   end
 
@@ -1176,8 +1175,7 @@ defmodule Code.Formatter.CallsTest do
       }
       """
 
-      # Doesn't preserve this because only the beginning has a newline
-      assert_format "call.{\nfoo, bar, baz}", "call.{foo, bar, baz}"
+      assert_format "call.{foo, bar, baz\n}", "call.{foo, bar, baz}"
     end
   end
 
