@@ -5,7 +5,15 @@ defmodule KernelTest do
 
   doctest Kernel
 
-  defp empty_list(), do: []
+  def id(arg), do: arg
+  def id(arg1, arg2), do: {arg1, arg2}
+  def empty_list(), do: []
+  def empty_map, do: %{}
+
+  defp purge(module) do
+    :code.delete(module)
+    :code.purge(module)
+  end
 
   defp assert_eval_raise(error, msg, string) do
     assert_raise error, msg, fn ->
@@ -13,6 +21,62 @@ defmodule KernelTest do
     end
   end
 
+  describe "=/2" do
+    test "can be reassigned" do
+      var = 1
+      id(var)
+      var = 2
+      assert var == 2
+    end
+
+    test "can be reassigned inside a list" do
+      _ = [var = 1, 2, 3]
+      id(var)
+      _ = [var = 2, 3, 4]
+      assert var == 2
+    end
+
+    test "can be reassigned inside a keyword list" do
+      _ = [a: var = 1, b: 2]
+      id(var)
+      _ = [b: var = 2, c: 3]
+      assert var == 2
+    end
+
+    test "can be reassigned inside a call" do
+      id(var = 1)
+      id(var)
+      id(var = 2)
+      assert var == 2
+    end
+
+    test "can be reassigned inside a multi-argument call" do
+      id(:arg, var = 1)
+      id(:arg, var)
+      id(:arg, var = 2)
+      assert var == 2
+
+      id(:arg, a: 1, b: var = 2)
+      id(:arg, var)
+      id(:arg, b: 2, c: var = 3)
+      assert var == 3
+    end
+
+    test "++/2 works in matches" do
+      [1, 2] ++ var = [1, 2]
+      assert var == []
+
+      [1, 2] ++ var = [1, 2, 3]
+      assert var == [3]
+
+      'ab' ++ var = 'abc'
+      assert var == 'c'
+
+      [:a, :b] ++ var = [:a, :b, :c]
+      assert var == [:c]
+    end
+  end
+
   test "=~/2" do
     assert "abcd" =~ ~r/c(d)/ == true
     assert "abcd" =~ ~r/e/ == false
@@ -81,6 +145,7 @@ defmodule KernelTest do
     end
   end
 
+  # Note we use `==` in assertions so `assert` does not rewrite `match?/2`.
   test "match?/2" do
     a = List.first([0])
     assert match?(b when b > a, 1) == true
@@ -88,6 +153,9 @@ defmodule KernelTest do
 
     assert match?(b when b > a, -1) == false
     assert binding() == [a: 0]
+
+    # Does not warn on underscored variables
+    assert match?(_unused, a) == true
   end
 
   def exported?, do: not_exported?()
@@ -243,6 +311,29 @@ defmodule KernelTest do
     assert_raise BadBooleanError, fn -> 0 or 1 end
   end
 
+  defp struct?(arg) when is_struct(arg), do: true
+  defp struct?(_arg), do: false
+
+  defp struct_or_map?(arg) when is_struct(arg) or is_map(arg), do: true
+  defp struct_or_map?(_arg), do: false
+
+  test "is_struct/1" do
+    assert is_struct(%{}) == false
+    assert is_struct([]) == false
+    assert is_struct(%Macro.Env{}) == true
+    assert is_struct(%{__struct__: "foo"}) == false
+    assert struct?(%Macro.Env{}) == true
+    assert struct?(%{__struct__: "foo"}) == false
+    assert struct?([]) == false
+    assert struct?(%{}) == false
+  end
+
+  test "is_struct/1 and other match works" do
+    assert struct_or_map?(%Macro.Env{}) == true
+    assert struct_or_map?(%{}) == true
+    assert struct_or_map?(10) == false
+  end
+
   test "if/2 boolean optimization does not leak variables during expansion" do
     if false do
       :ok
@@ -835,7 +926,7 @@ defmodule KernelTest do
       assert_raise KeyError, fn -> pop_in(users.bob[:age]) end
     end
 
-    test "pop_in/1/2 with nils" do
+    test "pop_in/1,2 with nils" do
       users = %{"john" => nil, "meg" => %{age: 23}}
       assert pop_in(users["john"][:age]) == {nil, %{"meg" => %{age: 23}}}
       assert pop_in(users, ["john", :age]) == {nil, %{"meg" => %{age: 23}}}
@@ -867,8 +958,6 @@ defmodule KernelTest do
       end
     end
 
-    def empty_map, do: %{}
-
     def by_index(index) do
       fn
         _, nil, next ->
@@ -1033,6 +1122,21 @@ defmodule KernelTest do
     end
   end
 
+  test "is_map_key/2" do
+    assert is_map_key(Map.new([]), :a) == false
+    assert is_map_key(Map.new(b: 1), :a) == false
+    assert is_map_key(Map.new(a: 1), :a) == true
+
+    assert_raise BadMapError, fn ->
+      is_map_key(empty_list(), :a)
+    end
+
+    case Map.new(a: 1) do
+      map when is_map_key(map, :a) -> true
+      _ -> flunk("invalid guard")
+    end
+  end
+
   test "tl/1" do
     assert tl([:one]) == []
     assert tl([1, 2, 3]) == [2, 3]
@@ -1111,9 +1215,4 @@ defmodule KernelTest do
       Code.eval_string(~s{~U[2015-01-13 13:00:07+00:30]})
     end
   end
-
-  defp purge(module) do
-    :code.delete(module)
-    :code.purge(module)
-  end
 end
