@@ -220,6 +220,8 @@ defmodule TaskTest do
       assert {{%RuntimeError{}, _}, {Task, :await, [^task, 5000]}} = catch_exit(Task.await(task))
     end
 
+    @compile {:no_warn_undefined, :module_does_not_exist}
+
     test "exits on task undef module error" do
       Process.flag(:trap_exit, true)
       task = Task.async(&:module_does_not_exist.undef/0)
@@ -229,6 +231,8 @@ defmodule TaskTest do
       assert {Task, :await, [^task, 5000]} = mfa
     end
 
+    @compile {:no_warn_undefined, {TaskTest, :undef, 0}}
+
     test "exits on task undef function error" do
       Process.flag(:trap_exit, true)
       task = Task.async(&TaskTest.undef/0)
@@ -580,6 +584,21 @@ defmodule TaskTest do
       [ok: :ok] = Task.async_stream([1], fn _ -> :ok end, timeout: :infinity) |> Enum.to_list()
     end
 
+    test "streams with fake down messages on the inbox" do
+      parent = self()
+
+      assert Task.async_stream([:ok], fn :ok ->
+               {:links, links} = Process.info(self(), :links)
+
+               for link <- links do
+                 send(link, {:DOWN, make_ref(), :process, parent, :oops})
+               end
+
+               :ok
+             end)
+             |> Enum.to_list() == [ok: :ok]
+    end
+
     test "with $callers" do
       grandparent = self()
 
