@@ -94,6 +94,12 @@ defmodule TypespecTest do
           def foo(var1, var2), do: {var1, var2}
         end
       end
+
+      assert_raise CompileError, ~r"type foo/0 undefined", fn ->
+        test_module do
+          @type omg :: __MODULE__.foo()
+        end
+      end
     end
 
     test "redefined type" do
@@ -948,43 +954,37 @@ defmodule TypespecTest do
       end
 
       assert [
-               {:type, {:"::", _, [{:type1, _, _}, {:boolean, _, _}]},
-                {TypespecTest.TypeModuleAttributes, _}}
+               {:type, {:"::", _, [{:type1, _, _}, {:boolean, _, _}]}, {TypeModuleAttributes, _}}
              ] = TypeModuleAttributes.type1()
 
       assert [
-               {:type, {:"::", _, [{:type3, _, _}, {:pid, _, _}]},
-                {TypespecTest.TypeModuleAttributes, _}},
-               {:type, {:"::", _, [{:type2, _, _}, {:atom, _, _}]},
-                {TypespecTest.TypeModuleAttributes, _}},
-               {:type, {:"::", _, [{:type1, _, _}, {:boolean, _, _}]},
-                {TypespecTest.TypeModuleAttributes, _}}
+               {:type, {:"::", _, [{:type3, _, _}, {:pid, _, _}]}, {TypeModuleAttributes, _}},
+               {:type, {:"::", _, [{:type2, _, _}, {:atom, _, _}]}, {TypeModuleAttributes, _}},
+               {:type, {:"::", _, [{:type1, _, _}, {:boolean, _, _}]}, {TypeModuleAttributes, _}}
              ] = TypeModuleAttributes.type2()
 
       assert [
                {:opaque, {:"::", _, [{:opaque1, _, _}, {:boolean, _, _}]},
-                {TypespecTest.TypeModuleAttributes, _}}
+                {TypeModuleAttributes, _}}
              ] = TypeModuleAttributes.opaque1()
 
       assert [
-               {:opaque, {:"::", _, [{:opaque3, _, _}, {:pid, _, _}]},
-                {TypespecTest.TypeModuleAttributes, _}},
+               {:opaque, {:"::", _, [{:opaque3, _, _}, {:pid, _, _}]}, {TypeModuleAttributes, _}},
                {:opaque, {:"::", _, [{:opaque2, _, _}, {:atom, _, _}]},
-                {TypespecTest.TypeModuleAttributes, _}},
+                {TypeModuleAttributes, _}},
                {:opaque, {:"::", _, [{:opaque1, _, _}, {:boolean, _, _}]},
-                {TypespecTest.TypeModuleAttributes, _}}
+                {TypeModuleAttributes, _}}
              ] = TypeModuleAttributes.opaque2()
 
       assert [
                {:typep, {:"::", _, [{:typep1, _, _}, {:boolean, _, _}]},
-                {TypespecTest.TypeModuleAttributes, _}}
+                {TypeModuleAttributes, _}}
              ] = TypeModuleAttributes.typep1()
 
       assert [
-               {:typep, {:"::", _, [{:typep2, _, _}, {:atom, _, _}]},
-                {TypespecTest.TypeModuleAttributes, _}},
+               {:typep, {:"::", _, [{:typep2, _, _}, {:atom, _, _}]}, {TypeModuleAttributes, _}},
                {:typep, {:"::", _, [{:typep1, _, _}, {:boolean, _, _}]},
-                {TypespecTest.TypeModuleAttributes, _}}
+                {TypeModuleAttributes, _}}
              ] = TypeModuleAttributes.typep2()
     after
       :code.delete(TypeModuleAttributes)
@@ -999,8 +999,8 @@ defmodule TypespecTest do
         @spec spec1 :: boolean
         def spec1, do: @spec
 
-        @callback callback2 :: boolean
-        @macrocallback macrocallback2 :: boolean
+        @callback callback2 :: var when var: boolean
+        @macrocallback macrocallback2 :: var when var: boolean
 
         @spec spec2 :: atom
         def spec2, do: @spec
@@ -1014,38 +1014,36 @@ defmodule TypespecTest do
       end
 
       assert [
-               {:spec, {:"::", _, [{:spec1, _, _}, {:boolean, _, _}]},
-                {TypespecTest.SpecModuleAttributes, _}}
+               {:spec, {:"::", _, [{:spec1, _, _}, {:boolean, _, _}]}, {SpecModuleAttributes, _}}
              ] = SpecModuleAttributes.spec1()
 
       assert [
-               {:spec, {:"::", _, [{:spec2, _, _}, {:atom, _, _}]},
-                {TypespecTest.SpecModuleAttributes, _}},
-               {:spec, {:"::", _, [{:spec1, _, _}, {:boolean, _, _}]},
-                {TypespecTest.SpecModuleAttributes, _}}
+               {:spec, {:"::", _, [{:spec2, _, _}, {:atom, _, _}]}, {SpecModuleAttributes, _}},
+               {:spec, {:"::", _, [{:spec1, _, _}, {:boolean, _, _}]}, {SpecModuleAttributes, _}}
              ] = SpecModuleAttributes.spec2()
 
       assert [
-               {:spec, {:"::", _, [{:spec3, _, _}, {:pid, _, _}]},
-                {TypespecTest.SpecModuleAttributes, _}},
-               {:spec, {:"::", _, [{:spec2, _, _}, {:atom, _, _}]},
-                {TypespecTest.SpecModuleAttributes, _}},
-               {:spec, {:"::", _, [{:spec1, _, _}, {:boolean, _, _}]},
-                {TypespecTest.SpecModuleAttributes, _}}
+               {:spec, {:"::", _, [{:spec3, _, _}, {:pid, _, _}]}, {SpecModuleAttributes, _}},
+               {:spec, {:"::", _, [{:spec2, _, _}, {:atom, _, _}]}, {SpecModuleAttributes, _}},
+               {:spec, {:"::", _, [{:spec1, _, _}, {:boolean, _, _}]}, {SpecModuleAttributes, _}}
              ] = SpecModuleAttributes.spec4()
 
       assert [
-               {:callback, {:"::", _, [{:callback2, _, _}, {:boolean, _, _}]},
-                {TypespecTest.SpecModuleAttributes, _}},
+               {:callback,
+                {:when, _,
+                 [{:"::", _, [{:callback2, _, _}, {:var, _, _}]}, [var: {:boolean, _, _}]]},
+                {SpecModuleAttributes, _}},
                {:callback, {:"::", _, [{:callback1, _, _}, {:integer, _, _}]},
-                {TypespecTest.SpecModuleAttributes, _}}
+                {SpecModuleAttributes, _}}
              ] = SpecModuleAttributes.callback()
 
       assert [
-               {:macrocallback, {:"::", _, [{:macrocallback2, _, _}, {:boolean, _, _}]},
-                {TypespecTest.SpecModuleAttributes, _}},
+               {:macrocallback,
+                {:when, _,
+                 [{:"::", _, [{:macrocallback2, _, _}, {:var, _, _}]}, [var: {:boolean, _, _}]]},
+                {SpecModuleAttributes, _}},
                {:macrocallback, {:"::", _, [{:macrocallback1, _, _}, {:integer, _, _}]},
-                {TypespecTest.SpecModuleAttributes, _}}
+                {SpecModuleAttributes, _}}
              ] = SpecModuleAttributes.macrocallback()
     after
       :code.delete(SpecModuleAttributes)
