@@ -128,8 +128,9 @@ defmodule ExUnit.Case do
         end
       end
 
-  If you are setting a `@moduletag`, you must set that after your
-  call to `use ExUnit.Case` otherwise you will see compilation errors.
+  If you are setting a `@moduletag` or `@describetag` attribute, you must
+  set them after your call to `use ExUnit.Case` otherwise you will see
+  compilation errors.
 
   If the same key is set via `@tag`, the `@tag` value has higher
   precedence.
@@ -154,7 +155,7 @@ defmodule ExUnit.Case do
     * `:timeout` - customizes the test timeout in milliseconds (defaults to 60000).
       Accepts `:infinity` as a timeout value.
 
-  The `:test_type` tag is automatically set by ExUnit, but is _not_ reserved.
+  The `:test_type` tag is automatically set by ExUnit, but is **not** reserved.
   This tag is available for users to customize if they desire.
 
   ## Filters
@@ -210,6 +211,8 @@ defmodule ExUnit.Case do
 
   """
 
+  @type env :: module() | Macro.Env.t()
+
   @reserved [:module, :file, :line, :test, :async, :registered, :describe]
 
   @doc false
@@ -220,14 +223,13 @@ defmodule ExUnit.Case do
     end
 
     quote do
-      async = !!unquote(opts)[:async]
+      unless Module.has_attribute?(__MODULE__, :ex_unit_tests) do
+        tag_check =
+          [:moduletag, :describetag, :tag]
+          |> Enum.any?(&Module.has_attribute?(__MODULE__, &1))
 
-      unless Module.get_attribute(__MODULE__, :ex_unit_tests) do
-        moduletag_check = Module.get_attribute(__MODULE__, :moduletag)
-        tag_check = Module.get_attribute(__MODULE__, :tag)
-
-        if moduletag_check || tag_check do
-          raise "you must set @tag and @moduletag after the call to \"use ExUnit.Case\""
+        if tag_check do
+          raise "you must set @tag, @describetag, and @moduletag after the call to \"use ExUnit.Case\""
         end
 
         attributes = [
@@ -235,7 +237,9 @@ defmodule ExUnit.Case do
           :tag,
           :describetag,
           :moduletag,
-          :ex_unit_registered,
+          :ex_unit_registered_test_attributes,
+          :ex_unit_registered_describe_attributes,
+          :ex_unit_registered_module_attributes,
           :ex_unit_used_describes
         ]
 
@@ -243,11 +247,17 @@ defmodule ExUnit.Case do
 
         @before_compile ExUnit.Case
         @after_compile ExUnit.Case
-        @ex_unit_async async
+        @ex_unit_async false
         @ex_unit_describe nil
         use ExUnit.Callbacks
       end
 
+      async = unquote(opts)[:async]
+
+      if is_boolean(async) do
+        @ex_unit_async async
+      end
+
       import ExUnit.Callbacks
       import ExUnit.Assertions
       import ExUnit.Case, only: [describe: 2, test: 1, test: 2, test: 3]
@@ -256,14 +266,11 @@ defmodule ExUnit.Case do
   end
 
   @doc """
-  Defines a test with a string.
+  Defines a test with `message`.
 
-  Provides a convenient macro that allows a test to be
-  defined with a string. This macro automatically inserts
-  the atom `:ok` as the last line of the test. That said,
-  a passing test always returns `:ok`, but, more importantly,
-  it forces Elixir to not tail call optimize the test and
-  therefore avoids hiding lines from the backtrace.
+  The test may also define a `var`, which will pattern match
+  on the test context. For more information on contexts, see
+  `ExUnit.Callbacks`.
 
   ## Examples
 
@@ -387,6 +394,10 @@ defmodule ExUnit.Case do
       after
         @ex_unit_describe nil
         Module.delete_attribute(__MODULE__, :describetag)
+
+        for attribute <- Module.get_attribute(__MODULE__, :ex_unit_registered_describe_attributes) do
+          Module.delete_attribute(__MODULE__, attribute)
+        end
       end
     end
   end
@@ -450,16 +461,25 @@ defmodule ExUnit.Case do
   pluralization.
   """
   def register_test(%{module: mod, file: file, line: line}, test_type, name, tags) do
-    moduletag = Module.get_attribute(mod, :moduletag)
-
-    unless moduletag do
+    unless Module.has_attribute?(mod, :ex_unit_tests) do
       raise "cannot define #{test_type}. Please make sure you have invoked " <>
               "\"use ExUnit.Case\" in the current module"
     end
 
-    registered_attributes = Module.get_attribute(mod, :ex_unit_registered)
-    registered = Map.new(registered_attributes, &{&1, Module.get_attribute(mod, &1)})
+    registered_attribute_keys = [
+      :ex_unit_registered_module_attributes,
+      :ex_unit_registered_describe_attributes,
+      :ex_unit_registered_test_attributes
+    ]
 
+    registered =
+      for key <- registered_attribute_keys,
+          attribute <- Module.get_attribute(mod, key),
+          into: %{} do
+        {attribute, Module.get_attribute(mod, attribute)}
+      end
+
+    moduletag = Module.get_attribute(mod, :moduletag)
     tag = Module.delete_attribute(mod, :tag)
     async = Module.get_attribute(mod, :ex_unit_async)
 
@@ -494,9 +514,9 @@ defmodule ExUnit.Case do
     test = %ExUnit.Test{name: name, case: mod, tags: tags, module: mod}
     Module.put_attribute(mod, :ex_unit_tests, test)
 
-    Enum.each(registered_attributes, fn attribute ->
+    for attribute <- Module.get_attribute(mod, :ex_unit_registered_test_attributes) do
       Module.delete_attribute(mod, attribute)
-    end)
+    end
 
     name
   end
@@ -505,11 +525,10 @@ defmodule ExUnit.Case do
   Registers a new attribute to be used during `ExUnit.Case` tests.
 
   The attribute values will be available through `context.registered`.
-  Registered values are cleared after each `ExUnit.Case.test/3` similar
+  Registered values are cleared after each `test/3` similar
   to `@tag`.
 
-  `Module.register_attribute/3` is used to register the attribute,
-  this function takes the same options.
+  This function takes the same options as `Module.register_attribute/3`.
 
   ## Examples
 
@@ -530,15 +549,121 @@ defmodule ExUnit.Case do
       end
 
   """
+  @spec register_attribute(env, atom, keyword) :: :ok
   def register_attribute(env, name, opts \\ [])
-
-  def register_attribute(%{module: module}, name, opts) do
-    register_attribute(module, name, opts)
-  end
+  def register_attribute(%{module: mod}, name, opts), do: register_attribute(mod, name, opts)
 
   def register_attribute(mod, name, opts) when is_atom(mod) and is_atom(name) and is_list(opts) do
+    register_attribute(:ex_unit_registered_test_attributes, mod, name, opts)
+  end
+
+  @doc """
+  Registers a new describe attribute to be used during `ExUnit.Case` tests.
+
+  The attribute values will be available through `context.registered`.
+  Registered values are cleared after each `describe/2` similar
+  to `@describetag`.
+
+  This function takes the same options as `Module.register_attribute/3`.
+
+  ## Examples
+
+      defmodule MyTest do
+        use ExUnit.Case
+
+        ExUnit.Case.register_describe_attribute(__MODULE__, :describe_fixtures, accumulate: true)
+
+        describe "using custom attribute" do
+          @describe_fixtures :user
+          @describe_fixtures {:post, insert: false}
+
+          test "has attribute", context do
+            assert context.registered.describe_fixtures == [{:post, insert: false}, :user]
+          end
+        end
+
+        describe "custom attributes are cleared per describe" do
+          test "doesn't have attributes", context do
+            assert context.registered.describe_fixtures == []
+          end
+        end
+      end
+
+  """
+  @doc since: "1.10.0"
+  @spec register_describe_attribute(env, atom, keyword) :: :ok
+  def register_describe_attribute(env, name, opts \\ [])
+
+  def register_describe_attribute(%{module: mod}, name, opts) do
+    register_describe_attribute(mod, name, opts)
+  end
+
+  def register_describe_attribute(mod, name, opts)
+      when is_atom(mod) and is_atom(name) and is_list(opts) do
+    register_attribute(:ex_unit_registered_describe_attributes, mod, name, opts)
+  end
+
+  @doc """
+  Registers a new module attribute to be used during `ExUnit.Case` tests.
+
+  The attribute values will be available through `context.registered`.
+
+  This function takes the same options as `Module.register_attribute/3`.
+
+  ## Examples
+
+      defmodule MyTest do
+        use ExUnit.Case
+
+        ExUnit.Case.register_module_attribute(__MODULE__, :module_fixtures, accumulate: true)
+
+        @module_fixtures :user
+        @module_fixtures {:post, insert: false}
+
+        test "using custom attribute", context do
+          assert context.registered.fixtures == [{:post, insert: false}, :user]
+        end
+
+        test "still using custom attribute", context do
+          assert context.registered.fixtures == [{:post, insert: false}, :user]
+        end
+      end
+
+  """
+  @doc since: "1.10.0"
+  @spec register_module_attribute(env, atom, keyword) :: :ok
+  def register_module_attribute(env, name, opts \\ [])
+
+  def register_module_attribute(%{module: mod}, name, opts) do
+    register_module_attribute(mod, name, opts)
+  end
+
+  def register_module_attribute(mod, name, opts)
+      when is_atom(mod) and is_atom(name) and is_list(opts) do
+    register_attribute(:ex_unit_registered_module_attributes, mod, name, opts)
+  end
+
+  defp register_attribute(type, mod, name, opts) do
+    validate_registered_attribute!(type, mod, name)
     Module.register_attribute(mod, name, opts)
-    Module.put_attribute(mod, :ex_unit_registered, name)
+    Module.put_attribute(mod, type, name)
+  end
+
+  defp validate_registered_attribute!(type, mod, name) do
+    registered_attribute_keys = [
+      :ex_unit_registered_module_attributes,
+      :ex_unit_registered_describe_attributes,
+      :ex_unit_registered_test_attributes
+    ]
+
+    for key <- registered_attribute_keys,
+        type != key and name in Module.get_attribute(mod, key) do
+      raise ArgumentError, "cannot register attribute #{inspect(name)} multiple times"
+    end
+
+    if Module.has_attribute?(mod, name) do
+      raise "you must set @#{name} after it has been registered"
+    end
   end
 
   defp validate_tags(tags) do
