@@ -61,12 +61,16 @@ defmodule ExUnit.DocTest do
       2
 
   If you don't want to assert for every result in a doctest, you can omit
-  the result:
+  the result. You can do so between expressions:
 
       iex> pid = spawn(fn -> :ok end)
       iex> is_pid(pid)
       true
 
+  As well as at the end:
+
+      iex> Mod.do_a_call_that_should_not_raise!(...)
+
   This is useful when the result is something variable (like a PID in the
   example above) or when the result is a complicated data structure and you
   don't want to show it all, but just parts of it or some of its properties.
@@ -291,9 +295,14 @@ defmodule ExUnit.DocTest do
     end) > 1
   end
 
+  defp test_case_content(expr, :test, location, stack, formatted) do
+    string_to_quoted(location, stack, expr, "\n" <> formatted) |> insert_assertions()
+  end
+
   defp test_case_content(expr, {:test, expected}, location, stack, formatted) do
-    expr_ast = string_to_quoted(location, stack, expr)
-    expected_ast = string_to_quoted(location, stack, expected)
+    doctest = "\n" <> formatted <> "\n" <> expected
+    expr_ast = string_to_quoted(location, stack, expr, doctest) |> insert_assertions()
+    expected_ast = string_to_quoted(location, stack, expected, doctest)
     last_expr_ast = last_expr(expr_ast)
 
     quote do
@@ -304,7 +313,7 @@ defmodule ExUnit.DocTest do
           :ok
 
         actual ->
-          doctest = unquote("\n" <> formatted <> "\n" <> expected)
+          doctest = unquote(doctest)
 
           expr =
             "#{unquote(Macro.to_string(last_expr_ast))} === #{unquote(String.trim(expected))}"
@@ -323,8 +332,9 @@ defmodule ExUnit.DocTest do
   end
 
   defp test_case_content(expr, {:inspect, expected}, location, stack, formatted) do
-    expr_ast = string_to_quoted(location, stack, expr)
-    expected_ast = string_to_quoted(location, stack, expected)
+    doctest = "\n" <> formatted <> "\n" <> expected
+    expr_ast = string_to_quoted(location, stack, expr, doctest) |> insert_assertions()
+    expected_ast = string_to_quoted(location, stack, expected, doctest)
     last_expr_ast = last_expr(expr_ast)
 
     quote do
@@ -335,7 +345,7 @@ defmodule ExUnit.DocTest do
           :ok
 
         actual ->
-          doctest = unquote("\n" <> formatted <> "\n" <> expected)
+          doctest = unquote(doctest)
 
           expr =
             "inspect(#{unquote(Macro.to_string(last_expr_ast))}) === " <>
@@ -355,15 +365,14 @@ defmodule ExUnit.DocTest do
   end
 
   defp test_case_content(expr, {:error, exception, message}, location, stack, formatted) do
-    expr_ast = string_to_quoted(location, stack, expr)
+    doctest = "\n" <> formatted <> "\n** (#{inspect(exception)}) #{inspect(message)}"
+    expr_ast = string_to_quoted(location, stack, expr, doctest)
 
     quote do
       stack = unquote(stack)
-      expected_exception = inspect(unquote(exception))
-
-      doctest =
-        unquote("\n" <> formatted <> "\n") <>
-          "** (#{expected_exception}) #{inspect(unquote(message))}"
+      exception = unquote(exception)
+      doctest = unquote(doctest)
+      message = unquote(message)
 
       try do
         unquote(expr_ast)
@@ -374,14 +383,14 @@ defmodule ExUnit.DocTest do
 
           message =
             cond do
-              actual_exception != unquote(exception) ->
-                "Doctest failed: expected exception #{expected_exception} but got " <>
+              actual_exception != exception ->
+                "Doctest failed: expected exception #{inspect(exception)} but got " <>
                   "#{inspect(actual_exception)} with message #{inspect(actual_message)}"
 
-              actual_message != unquote(message) ->
+              actual_message != message ->
                 "Doctest failed: wrong message for #{inspect(actual_exception)}\n" <>
                   "expected:\n" <>
-                  "  #{inspect(unquote(message))}\n" <>
+                  "  #{inspect(message)}\n" <>
                   "actual:\n" <> "  #{inspect(actual_message)}"
 
               true ->
@@ -394,7 +403,7 @@ defmodule ExUnit.DocTest do
       else
         _ ->
           message =
-            "Doctest failed: expected exception #{expected_exception} but nothing was raised"
+            "Doctest failed: expected exception #{inspect(exception)} but nothing was raised"
 
           error = [message: message, doctest: doctest]
           reraise ExUnit.AssertionError, error, stack
@@ -405,7 +414,7 @@ defmodule ExUnit.DocTest do
   defp test_import(_mod, false), do: []
   defp test_import(mod, _), do: [quote(do: import(unquote(mod)))]
 
-  defp string_to_quoted(location, stack, expr) do
+  defp string_to_quoted(location, stack, expr, doctest) do
     try do
       Code.string_to_quoted!(expr, location)
     rescue
@@ -427,8 +436,8 @@ defmodule ExUnit.DocTest do
           end
 
         opts =
-          if String.valid?(expr) do
-            [message: message, expr: String.trim(expr)]
+          if String.valid?(doctest) do
+            [message: message, doctest: doctest]
           else
             [message: message]
           end
@@ -502,10 +511,6 @@ defmodule ExUnit.DocTest do
     adjust_indent(:text, lines, line_no, [], 0, module)
   end
 
-  defp adjust_indent(:after_prompt, [], line_no, _adjusted_lines, _indent, module) do
-    raise_incomplete_doctest(line_no, module)
-  end
-
   defp adjust_indent(_kind, [], _line_no, adjusted_lines, _indent, _module) do
     Enum.reverse(adjusted_lines)
   end
@@ -527,7 +532,7 @@ defmodule ExUnit.DocTest do
 
     case String.trim_leading(line) do
       "" ->
-        raise_incomplete_doctest(line_no, module)
+        :ok
 
       ^stripped_line ->
         :ok
@@ -811,6 +816,9 @@ defmodule ExUnit.DocTest do
 
   defp tag_expected(string) do
     case string do
+      "" ->
+        :test
+
       "** (" <> error ->
         [mod, message] = :binary.split(error, ")")
         {:error, Module.concat([mod]), String.trim_leading(message)}
@@ -843,10 +851,25 @@ defmodule ExUnit.DocTest do
   defp last_expr({:__block__, _, [_ | _] = block}), do: block |> List.last() |> last_expr()
   defp last_expr(other), do: other
 
-  defp raise_incomplete_doctest(line_no, module) do
-    raise Error,
-      line: line_no,
-      module: module,
-      message: "expected non-blank line to follow iex> prompt"
+  defp insert_assertions({:__block__, meta, block}),
+    do: {:__block__, meta, Enum.map(block, &insert_match_assertion/1)}
+
+  defp insert_assertions(ast),
+    do: insert_match_assertion(ast)
+
+  defp insert_match_assertion({:=, _, [{var, _, context}, _]} = ast)
+       when is_atom(var) and is_atom(context),
+       do: ast
+
+  defp insert_match_assertion({:=, meta, [left, right]}),
+    do: {{:., meta, [__MODULE__, :__assert__]}, meta, [{:=, meta, [left, right]}]}
+
+  defp insert_match_assertion(ast),
+    do: ast
+
+  @doc false
+  defmacro __assert__({:=, _, [left, right]} = assertion) do
+    code = Macro.escape(assertion, prune_metadata: true)
+    ExUnit.Assertions.__match__(left, right, code, :ok, __CALLER__)
   end
 end
