@@ -16,10 +16,8 @@ defmodule Logger do
       performant when required but also apply backpressure
       when under stress.
 
-    * Plugs into Erlang's [`:logger`](http://erlang.org/doc/man/logger.html)
-      (from Erlang/OTP 21) to convert terms to Elixir syntax or wraps
-      Erlang's [`:error_logger`](http://erlang.org/doc/man/error_logger.html)
-      in earlier Erlang/OTP versions to prevent it from overflowing.
+    * Integrates with Erlang's [`:logger`](http://erlang.org/doc/man/logger.html)
+      to convert terms to Elixir syntax.
 
   Logging is useful for tracking when an event of interest happens in your
   system. For example, it may be helpful to log whenever a user is deleted.
@@ -52,12 +50,55 @@ defmodule Logger do
     * `:info` - for information of any kind
     * `:warn` - for warnings
     * `:error` - for errors
-    
-  For example, `:info` takes precedence over `:debug`. If your log 
-  level is set to `:info`, `:info`, `:warn`, and `:error` will be 
-  printed to the console. If your log level is set to `:warn`, only 
+
+  For example, `:info` takes precedence over `:debug`. If your log
+  level is set to `:info`, `:info`, `:warn`, and `:error` will be
+  printed to the console. If your log level is set to `:warn`, only
   `:warn` and `:error` will be printed.
 
+  ## Metadata
+
+  Whenever a message is logged, additional information can be given
+  via metadata. Each log operation, such as `Logger.info/2`, allows
+  metadata to be given as argument.
+
+  Furthermore, metadata can be set per process with `Logger.metadata/1`.
+
+  Some metadata, however, is always added automatically by Logger
+  whenever possible. Those are:
+
+    * `:application` - the current application
+
+    * `:mfa` - the current module, function and arity
+
+    * `:file` - the current file
+
+    * `:line` - the current line
+
+    * `:pid` - the current process identifier
+
+    * `:initial_call` - the initial call that started the process
+
+    * `:registered_name` - the process registered name as an atom
+
+    * `:domain` - a list of domains for the logged message. For example,
+      all Elixir reports default to `[:elixir]`. Erlang reports may start
+      with `[:otp]` or `[:sasl]`
+
+    * `:crash_reason` - a two-element tuple with the throw/error/exit reason
+      as first argument and the stacktrace as second. A throw will always be
+      `{:nocatch, term}`. An error is always an `Exception` struct. All other
+      entries are exits. The console backend ignores this metadata by default
+      but it can be useful to other backends, such as the ones that report
+      errors to third-party services
+
+  Note that all metadata is optional and may not always be available.
+  The `:mfa`, `:file`, `:line`, and similar metadata are automatically
+  included when using `Logger` macros. `Logger.bare_log/3` does not include
+  any metadata beyond the `:pid` by default. Other metadata, such as
+  `:crash_reason`, `:initial_call`, and `:registered_name` are available
+  only inside behaviours such as GenServer, Supervisor, and others.
+
   ## Configuration
 
   `Logger` supports a wide range of configurations.
@@ -83,19 +124,25 @@ defmodule Logger do
 
     * `:compile_time_application` - sets the `:application` metadata value
       to the configured value at compilation time. This configuration is
-      usually only useful for build tools to automatically add the
-      application to the metadata for `Logger.debug/2`, `Logger.info/2`, etc.
-      style of calls.
+      automatically set by Mix and made available as metadata when logging.
 
     * `:compile_time_purge_matching` - purges *at compilation time* all calls
       that match the given conditions. This means that `Logger` calls with
       level lower than this option will be completely removed at compile time,
       accruing no overhead at runtime. This configuration expects a list of
       keyword lists. Each keyword list contains a metadata key and the matching
-      value that should be purged. A special key named `:level_lower_than` can
-      be used to purge all messages with a lower logger level. Remember that
-      if you want to purge log calls from a dependency, the dependency must be
-      recompiled.
+      value that should be purged. Some special keys are supported:
+
+        * `:level_lower_than` - purges all messages with a lower logger level
+        * `:module` - purges all messages with the matching module
+        * `:function` - purges all messages with the "function/arity"
+
+      Remember that if you want to purge log calls from a dependency, the
+      dependency must be recompiled.
+
+    * `:start_options` - passes start options to Logger's main process, such
+      as `:spawn_opt` and `:hibernate_after`. All options in `t:GenServer.option`
+      are accepted, except `:name`.
 
   For example, to configure the `:backends` and purge all calls that happen
   at compile time with level lower than `:info` in a `config/config.exs` file:
@@ -168,11 +215,14 @@ defmodule Logger do
         level: :warn,
         truncate: 4096
 
-  ### Error logger configuration
+  ### Erlang/OTP integration
 
-  The following configuration applies to `Logger`'s wrapper around
-  Erlang's logging functionalities. All the configurations below must
-  be set before the `:logger` application starts.
+  From Elixir v1.10, Elixir's Logger is fully integrated with Erlang's
+  logger. They share the same `Logger.level/0`, any metadata set with
+  `Logger.metadata/1` applies to both, and so on.
+
+  Elixir also supports formatting Erlang reports using Elixir syntax.
+  This can be controlled with two configurations:
 
     * `:handle_otp_reports` - redirects OTP reports to `Logger` so
       they are formatted in Elixir terms. This effectively disables
@@ -183,22 +233,8 @@ defmodule Logger do
       terms. Your application must guarantee `:sasl` is started before
       `:logger`. This means you may see some initial reports written
       in Erlang syntax until the Logger application kicks in.
-      Defaults to `false`.
-
-  From Erlang/OTP 21, `:handle_sasl_reports` only has an effect if
-  `:handle_otp_reports` is true.
-
-  The following configurations apply only for Erlang/OTP 20 and earlier:
-
-    * `:discard_threshold_for_error_logger` - if `:error_logger` has more than
-      `discard_threshold` messages in its inbox, messages will be dropped
-      until the message queue goes down to `discard_threshold * 0.75`
-      entries. The threshold will be checked once again after 10% of threshold
-      messages are processed, to avoid messages from being constantly dropped.
-      For example, if the threshold is 500 (the default) and the inbox has
-      600 messages, 225 messages will dropped, bringing the inbox down to
-      375 (0.75 * threshold) entries and 50 (0.1 * threshold) messages will
-      be processed before the threshold is checked once again.
+      Defaults to `false`. This option only has an effect if
+      `:handle_otp_reports` is true.
 
   For example, to configure `Logger` to redirect all Erlang messages using a
   `config/config.exs` file:
@@ -225,11 +261,12 @@ defmodule Logger do
 
   The initial backends are loaded via the `:backends` configuration,
   which must be set before the `:logger` application is started.
+  Backends can also be added dynamically through `add_backend/2`.
 
   ### Console backend
 
-  The console backend logs messages by printing them to the console. It supports
-  the following options:
+  The console backend logs messages by printing them to the console.
+  It supports the following options:
 
     * `:level` - the level to be logged by this backend.
       Note that messages are filtered by the general
@@ -278,55 +315,21 @@ defmodule Logger do
         format: "\n$time $metadata[$level] $levelpad$message\n",
         metadata: [:user_id]
 
-  ## Metadata
-
-  In addition to the keys provided by the user via `Logger.metadata/1`,
-  the following extra keys are available to the `:metadata` list:
-
-    * `:application` - the current application
-
-    * `:module` - the current module
-
-    * `:function` - the current function
-
-    * `:file` - the current file
-
-    * `:line` - the current line
-
-    * `:pid` - the current process identifier
-
-    * `:crash_reason` - a two-element tuple with the throw/error/exit reason
-      as first argument and the stacktrace as second. A throw will always be
-      `{:nocatch, term}`. An error is always an `Exception` struct. All other
-      entries are exits. The console backend ignores this metadata by default
-      but it can be useful to other backends, such as the ones that report
-      errors to third-party services
-
-    * `:initial_call` - the initial call that started the process
-
-    * `:registered_name` - the process registered name as an atom
-
-  Note that all metadata is optional and may not always be available.
-  The `:module`, `:function`, `:line`, and similar metadata are automatically
-  included when using `Logger` macros. `Logger.bare_log/3` does not include
-  any metadata beyond the `:pid` by default. Other metadata, such as
-  `:crash_reason`, `:initial_call`, and `:registered_name` are extracted
-  from Erlang/OTP crash reports and available only in those cases.
-
   ### Custom formatting
 
-  The console backend allows you to customize the format of your log messages
-  with the `:format` option.
+  The console backend allows you to customize the format of your
+  log messages with the `:format` option.
 
-  You may set `:format` to either a string or a `{module, function}` tuple if
-  you wish to provide your own format function. Here is an example of how to
-  configure the `:console` backend in a `config/config.exs` file:
+  You may set `:format` to either a string or a `{module, function}`
+  tuple if you wish to provide your own format function. Here is an
+  example of how to configure the `:console` backend in a
+  `config/config.exs` file:
 
       config :logger, :console,
         format: {MyConsoleLogger, :format}
 
-  And here is an example of how you can define `MyConsoleLogger.format/4` from the
-  above configuration:
+  And here is an example of how you can define `MyConsoleLogger.format/4`
+  from the above configuration:
 
       defmodule MyConsoleLogger do
         def format(level, message, timestamp, metadata) do
@@ -334,10 +337,10 @@ defmodule Logger do
         end
       end
 
-  It is extremely important that **the formatting function does not fail**, as
-  it will bring that particular logger instance down, causing your system to
-  temporarily lose messages. If necessary, wrap the function in a `rescue` and
-  log a default message instead:
+  It is extremely important that **the formatting function does
+  not fail**, as it will bring that particular logger instance down,
+  causing your system to temporarily lose messages. If necessary,
+  wrap the function in a `rescue` and log a default message instead:
 
       defmodule MyConsoleLogger do
         def format(level, message, timestamp, metadata) do
@@ -350,94 +353,144 @@ defmodule Logger do
   The `{module, function}` will be invoked with four arguments:
 
     * the log level: an atom
-    * the message: this is usually chardata, but in some cases it may not be.
-      Since the formatting function should *never* fail, you need to prepare for
-      the message being anything (and do something like the `rescue` in the example
-      above)
+    * the message: this is usually chardata, but in some cases it
+      may contain invalid data. Since the formatting function should
+      *never* fail, you need to prepare for the message being anything
     * the current timestamp: a term of type `t:Logger.Formatter.time/0`
     * the metadata: a keyword list
 
-  You can read more about formatting in `Logger.Formatter`.
+  You can read more about formatting in `Logger.Formatter`, especially
+  if you want to support custom formatting in a custom backend.
 
   ### Custom backends
 
-  Any developer can create their own `Logger` backend.
-  Since `Logger` is an event manager powered by `:gen_event`,
-  writing a new backend is a matter of creating an event
-  handler, as described in the [`:gen_event`](http://erlang.org/doc/man/gen_event.html)
-  documentation.
+  Any developer can create their own `Logger` backend. Since `Logger`
+  is an event manager powered by `:gen_event`, writing a new backend
+  is a matter of creating an event handler, as described in the
+  [`:gen_event`](http://erlang.org/doc/man/gen_event.html) documentation.
 
   From now on, we will be using the term "event handler" to refer
   to your custom backend, as we head into implementation details.
 
-  Once the `:logger` application starts, it installs all event handlers listed under
-  the `:backends` configuration into the `Logger` event manager. The event
-  manager and all added event handlers are automatically supervised by `Logger`.
+  Once the `:logger` application starts, it installs all event handlers
+  listed under the `:backends` configuration into the `Logger` event
+  manager. The event manager and all added event handlers are automatically
+  supervised by `Logger`.
 
-  Once initialized, the handler should be designed to handle events
-  in the following format:
+  Note that if a backend fails to start by returning `{:error, :ignore}`
+  from its `init/1` callback, then it's not added to the backends but
+  nothing fails. If a backend fails to start by returning `{:error, reason}`
+  from its `init/1` callback, the `:logger` application will fail to start.
 
-      {level, group_leader, {Logger, message, timestamp, metadata}} | :flush
+  Once initialized, the handler should be designed to handle the
+  following events:
 
-  where:
+    * `{level, group_leader, {Logger, message, timestamp, metadata}}` where:
+      * `level` is one of `:debug`, `:info`, `:warn`, or `:error`, as previously
+        described
+      * `group_leader` is the group leader of the process which logged the message
+      * `{Logger, message, timestamp, metadata}` is a tuple containing information
+        about the logged message:
+        * the first element is always the atom `Logger`
+        * `message` is the actual message (as chardata)
+        * `timestamp` is the timestamp for when the message was logged, as a
+          `{{year, month, day}, {hour, minute, second, millisecond}}` tuple
+        * `metadata` is a keyword list of metadata used when logging the message
 
-    * `level` is one of `:debug`, `:info`, `:warn`, or `:error`, as previously
-      described
-    * `group_leader` is the group leader of the process which logged the message
-    * `{Logger, message, timestamp, metadata}` is a tuple containing information
-      about the logged message:
-      * the first element is always the atom `Logger`
-      * `message` is the actual message (as chardata)
-      * `timestamp` is the timestamp for when the message was logged, as a
-        `{{year, month, day}, {hour, minute, second, millisecond}}` tuple
-      * `metadata` is a keyword list of metadata used when logging the message
+    * `:flush`
 
-  It is recommended that handlers ignore messages where
-  the group leader is in a different node than the one where
-  the handler is installed. For example:
+  It is recommended that handlers ignore messages where the group
+  leader is in a different node than the one where the handler is
+  installed. For example:
 
       def handle_event({_level, gl, {Logger, _, _, _}}, state)
           when node(gl) != node() do
         {:ok, state}
       end
 
-  In the case of the event `:flush` handlers should flush any pending data. This
-  event is triggered by `flush/0`.
+  In the case of the event `:flush` handlers should flush any pending
+  data. This event is triggered by `Logger.flush/0`.
 
-  Furthermore, backends can be configured via the
-  `configure_backend/2` function which requires event handlers
-  to handle calls of the following format:
+  Furthermore, backends can be configured via the `configure_backend/2`
+  function which requires event handlers to handle calls of the
+  following format:
 
       {:configure, options}
 
-  where `options` is a keyword list. The result of the call is
-  the result returned by `configure_backend/2`. The recommended
-  return value for successful configuration is `:ok`.
+  where `options` is a keyword list. The result of the call is the result
+  returned by `configure_backend/2`. The recommended return value for
+  successful configuration is `:ok`. For example:
 
-  It is recommended that backends support at least the following
-  configuration options:
+      def handle_call({:configure, options}, state) do
+        new_state = reconfigure_state(state, options)
+        {:ok, :ok, new_state}
+      end
+
+  It is recommended that backends support at least the following configuration
+  options:
 
     * `:level` - the logging level for that backend
     * `:format` - the logging format for that backend
     * `:metadata` - the metadata to include in that backend
 
-  Check the implementation for `Logger.Backends.Console`, for
-  examples on how to handle the recommendations in this section
-  and how to process the existing options.
+  Check `Logger.Backends.Console`'s implementation for examples on how
+  to handle the recommendations in this section and how to process the
+  existing options.
+
+  ### Erlang/OTP handlers
+
+  While Elixir Logger provides backends, Erlang/OTP logger provides handlers.
+  They represent the same concept: the ability to integrate into the logging
+  system to handle each logged message/event.
+
+  However, implementation-wise, they have the following differences:
+
+    * Elixir backends run in a separate process which comes with overload
+      protection. However, because this process is a single GenEvent, any
+      long running action should be avoided, as it can lead to bottlenecks
+      in the system
+
+    * Erlang handlers run in the same process as the process logging the
+      message/event. This gives developers more flexibility but they should
+      avoid perform any long running action in such handlers, as it may
+      slow down the action being executed considerably. At the moment, there
+      is no built-in overload protection for Erlang handlers, so it is your
+      responsibility to implement it
+
+  The good news is that developers can use third-party implementations of
+  both Elixir backends and Erlang handlers.
+
+  Elixir backends can be configured directly under the `:logger` application
+  in your `config/config.exs`:
+
+      config :logger, backends: [ACustomBackend]
+
+  Erlang/OTP handlers must be listed under your own application:
+
+      config :my_app, :logger,
+        [:handler, :name_of_the_handler, ACustomHandler, configuration = %{}]
+
+  And then explicitly attached in your `c:Application.start/2` callback:
+
+      :logger.add_handlers(:my_app)
+
+  Note we do not recommend configuring Erlang/OTP's logger directly under
+  the `:kernel` application in your `config/config.exs`, like this:
+
+      config :kernel, :logger, ...
+
+  This is because by the time Elixir starts, Erlang's kernel has already
+  been started, which means the configuration above would have no effect.
   """
 
+  @type level :: :error | :warn | :info | :debug
   @type backend :: :gen_event.handler()
   @type message :: IO.chardata() | String.Chars.t()
-  @type level :: :error | :info | :warn | :debug
   @type metadata :: keyword()
-  @levels [:error, :info, :warn, :debug]
+  @levels [:error, :warn, :info, :debug]
 
-  @metadata :logger_metadata
-  @compile {:inline, __metadata__: 0}
-
-  defp __metadata__ do
-    Process.get(@metadata) || {true, []}
-  end
+  @metadata :logger_enabled
+  @compile {:inline, enabled?: 1}
 
   @doc """
   Alters the current process metadata according the given keyword list.
@@ -448,24 +501,19 @@ defmodule Logger do
   """
   @spec metadata(metadata) :: :ok
   def metadata(keyword) do
-    {enabled?, metadata} = __metadata__()
-    Process.put(@metadata, {enabled?, into_metadata(keyword, metadata)})
-    :ok
-  end
+    case :logger.get_process_metadata() do
+      :undefined ->
+        reset_metadata(keyword)
 
-  defp into_metadata([], metadata), do: metadata
-  defp into_metadata(keyword, metadata), do: into_metadata(keyword, [], metadata)
+      map when is_map(map) ->
+        metadata =
+          Enum.reduce(keyword, map, fn
+            {k, nil}, acc -> Map.delete(acc, k)
+            {k, v}, acc -> Map.put(acc, k, v)
+          end)
 
-  defp into_metadata([{key, nil} | keyword], prepend, metadata) do
-    into_metadata(keyword, prepend, :lists.keydelete(key, 1, metadata))
-  end
-
-  defp into_metadata([{key, _} = pair | keyword], prepend, metadata) do
-    into_metadata(keyword, [pair | prepend], :lists.keydelete(key, 1, metadata))
-  end
-
-  defp into_metadata([], prepend, metadata) do
-    prepend ++ metadata
+        :ok = :logger.set_process_metadata(metadata)
+    end
   end
 
   @doc """
@@ -473,17 +521,22 @@ defmodule Logger do
   """
   @spec metadata() :: metadata
   def metadata() do
-    __metadata__() |> elem(1)
+    case :logger.get_process_metadata() do
+      :undefined -> []
+      map when is_map(map) -> Map.to_list(map)
+    end
   end
 
   @doc """
   Resets the current process metadata to the given keyword list.
   """
   @spec reset_metadata(metadata) :: :ok
-  def reset_metadata(keywords \\ []) do
-    {enabled?, _metadata} = __metadata__()
-    Process.put(@metadata, {enabled?, []})
-    metadata(keywords)
+  def reset_metadata(keyword \\ []) do
+    :ok = :logger.set_process_metadata(filter_out_nils(keyword))
+  end
+
+  defp filter_out_nils(keyword) do
+    for {_k, v} = elem <- keyword, v != nil, into: %{}, do: elem
   end
 
   @doc """
@@ -493,7 +546,7 @@ defmodule Logger do
   """
   @spec enable(pid) :: :ok
   def enable(pid) when pid == self() do
-    Process.put(@metadata, {true, metadata()})
+    Process.delete(@metadata)
     :ok
   end
 
@@ -504,17 +557,31 @@ defmodule Logger do
   """
   @spec disable(pid) :: :ok
   def disable(pid) when pid == self() do
-    Process.put(@metadata, {false, metadata()})
+    Process.put(@metadata, false)
     :ok
   end
 
+  @doc """
+  Returns whether the logging is enabled for given process.
+
+  Currently the only accepted PID is `self()`.
+  """
+  @doc since: "1.10.0"
+  @spec enabled?(pid) :: boolean
+  def enabled?(pid) when pid == self() do
+    Process.get(@metadata, true)
+  end
+
   @doc """
   Retrieves the `Logger` level.
 
   The `Logger` level can be changed via `configure/1`.
   """
-  @spec level() :: level
-  defdelegate level(), to: Logger.Config
+  @spec level() :: level()
+  def level() do
+    %{level: level} = :logger.get_primary_config()
+    Logger.Handler.erlang_level_to_elixir_level(level)
+  end
 
   @doc """
   Compares log levels.
@@ -535,7 +602,12 @@ defmodule Logger do
 
   """
   @spec compare_levels(level, level) :: :lt | :eq | :gt
-  defdelegate compare_levels(left, right), to: Logger.Config
+  def compare_levels(left, right) do
+    :logger.compare_levels(
+      Logger.Handler.elixir_level_to_erlang_level(left),
+      Logger.Handler.elixir_level_to_erlang_level(right)
+    )
+  end
 
   @doc """
   Configures the logger.
@@ -558,7 +630,13 @@ defmodule Logger do
   ]
   @spec configure(keyword) :: :ok
   def configure(options) do
-    Logger.Config.configure(Keyword.take(options, @valid_options))
+    options = Keyword.take(options, @valid_options)
+
+    # We serialize the writes
+    Logger.Config.configure(options)
+
+    # Then we can read from the writes
+    :ok = :logger.set_handler_config(Logger, %{config: %{}})
   end
 
   @doc """
@@ -570,23 +648,43 @@ defmodule Logger do
   """
   @spec flush :: :ok
   def flush do
-    _ = Process.whereis(:error_logger) && :gen_event.which_handlers(:error_logger)
     :gen_event.sync_notify(Logger, :flush)
   end
 
   @doc """
   Adds a new backend.
 
+  Adding a backend calls the `init/1` function in that backend
+  with the name of the backend as its argument. For example,
+  calling
+
+      Logger.add_backend(MyBackend)
+
+  will call `MyBackend.init(MyBackend)` to initialize the new
+  backend. If the backend's `init/1` callback returns `{:ok, _}`,
+  then this function returns `{:ok, pid}`. If the handler returns
+  `{:error, :ignore}` from `init/1`, this function still returns
+  `{:ok, pid}` but the handler is not started. If the handler
+  returns `{:error, reason}` from `init/1`, this function returns
+  `{:error, {reason, info}}` where `info` is more information on
+  the backend that failed to start.
+
   Backends added by this function are not persisted. Therefore
   if the Logger application or supervision tree is restarted,
   the backend won't be available. If you need this guarantee,
-  then configure the backend via the application environment.
+  then configure the backend via the application environment:
+
+      config :logger, :backends, [MyBackend]
 
   ## Options
 
     * `:flush` - when `true`, guarantees all messages currently sent
       to `Logger` are processed before the backend is added
 
+  ## Examples
+
+      {:ok, _pid} = Logger.add_backend(MyBackend, flush: true)
+
   """
   @spec add_backend(backend, keyword) :: Supervisor.on_start_child()
   def add_backend(backend, opts \\ []) do
@@ -655,65 +753,62 @@ defmodule Logger do
   anonymous functions to `bare_log/3` and they will only be evaluated
   if there is something to be logged.
   """
-  @spec bare_log(level, message | (() -> message | {message, keyword}), keyword) ::
-          :ok | {:error, :noproc} | {:error, term}
+  @spec bare_log(level, message | (() -> message | {message, keyword}), keyword) :: :ok
   def bare_log(level, chardata_or_fun, metadata \\ []) do
-    case __should_log__(level) do
-      :error -> :ok
-      info -> __do_log__(info, chardata_or_fun, metadata)
+    case __should_log__(level, nil) do
+      nil -> :ok
+      level -> __do_log__(level, chardata_or_fun, Map.new(metadata))
     end
   end
 
   @doc false
-  def __should_log__(level) when level in @levels do
-    case __metadata__() do
-      {true, pdict} ->
-        case Logger.Config.log_data(level) do
-          {:discard, _config} -> :error
-          {mode, config} -> {level, mode, config, pdict}
-        end
+  def __should_log__(level, module) when level in @levels do
+    level = Logger.Handler.elixir_level_to_erlang_level(level)
 
-      {false, _} ->
-        :error
+    if enabled?(self()) and :logger.allow(level, module) do
+      level
     end
   end
 
   @doc false
-  def __do_log__({level, mode, config, pdict}, chardata_or_fun, metadata)
-      when is_list(metadata) do
-    %{utc_log: utc_log?, truncate: truncate} = config
-    metadata = [pid: self()] ++ into_metadata(metadata, pdict)
+  def __do_log__(level, fun, metadata) when is_function(fun, 0) and is_map(metadata) do
+    case fun.() do
+      {msg, meta} ->
+        :logger.macro_log(%{}, level, msg, Enum.into(meta, add_elixir_domain(metadata)))
 
-    case normalize_message(chardata_or_fun, metadata) do
-      {message, metadata} ->
-        tuple = {Logger, truncate(message, truncate), Logger.Utils.timestamp(utc_log?), metadata}
-
-        try do
-          notify(mode, {level, Process.group_leader(), tuple})
-          :ok
-        rescue
-          ArgumentError -> {:error, :noproc}
-        catch
-          :exit, reason -> {:error, reason}
-        end
-
-      :skip ->
-        :ok
+      msg when is_binary(msg) or is_list(msg) ->
+        :logger.macro_log(%{}, level, msg, add_elixir_domain(metadata))
     end
   end
 
+  def __do_log__(level, chardata, metadata)
+      when (is_binary(chardata) or is_list(chardata)) and is_map(metadata) do
+    :logger.macro_log(%{}, level, chardata, add_elixir_domain(metadata))
+  end
+
+  # # TODO: Remove that in Elixir 2.0
+  def __do_log__(level, other, metadata) do
+    IO.warn("passing #{inspect(other)} to Logger is deprecated, expected a binary or an iolist")
+    :logger.macro_log(%{}, level, to_string(other), add_elixir_domain(metadata))
+  end
+
+  defp add_elixir_domain(%{domain: domain} = metadata) when is_list(domain) do
+    %{metadata | domain: [:elixir | domain]}
+  end
+
+  defp add_elixir_domain(metadata), do: Map.put(metadata, :domain, [:elixir])
+
   @doc """
   Logs a warning message.
 
-  Returns `:ok` or an `{:error, reason}` tuple.
+  Returns `:ok`.
 
   ## Examples
 
       Logger.warn("knob turned too far to the right")
-      Logger.warn(fn -> "dynamically calculated warning" end)
-      Logger.warn(fn -> {"dynamically calculated warning", [additional: :metadata]} end)
 
   """
+  # TODO: Deprecate it in favour of `warning/1-2` macro
   defmacro warn(chardata_or_fun, metadata \\ []) do
     maybe_log(:warn, chardata_or_fun, metadata, __CALLER__)
   end
@@ -721,13 +816,11 @@ defmodule Logger do
   @doc """
   Logs an info message.
 
-  Returns `:ok` or an `{:error, reason}` tuple.
+  Returns `:ok`.
 
   ## Examples
 
       Logger.info("mission accomplished")
-      Logger.info(fn -> "dynamically calculated info" end)
-      Logger.info(fn -> {"dynamically calculated info", [additional: :metadata]} end)
 
   """
   defmacro info(chardata_or_fun, metadata \\ []) do
@@ -737,13 +830,11 @@ defmodule Logger do
   @doc """
   Logs an error message.
 
-  Returns `:ok` or an `{:error, reason}` tuple.
+  Returns `:ok`.
 
   ## Examples
 
       Logger.error("oops")
-      Logger.error(fn -> "dynamically calculated error" end)
-      Logger.error(fn -> {"dynamically calculated error", [additional: :metadata]} end)
 
   """
   defmacro error(chardata_or_fun, metadata \\ []) do
@@ -753,13 +844,11 @@ defmodule Logger do
   @doc """
   Logs a debug message.
 
-  Returns `:ok` or an `{:error, reason}` tuple.
+  Returns `:ok`.
 
   ## Examples
 
       Logger.debug("hello?")
-      Logger.debug(fn -> "dynamically calculated debug" end)
-      Logger.debug(fn -> {"dynamically calculated debug", [additional: :metadata]} end)
 
   """
   defmacro debug(chardata_or_fun, metadata \\ []) do
@@ -769,7 +858,7 @@ defmodule Logger do
   @doc """
   Logs a message with the given `level`.
 
-  Returns `:ok` or an `{:error, reason}` tuple.
+  Returns `:ok`.
 
   The macros `debug/2`, `warn/2`, `info/2`, and `error/2` are
   preferred over this macro as they can automatically eliminate
@@ -781,20 +870,24 @@ defmodule Logger do
   end
 
   defp macro_log(level, data, metadata, caller) do
-    %{module: module, function: fun, file: file, line: line} = caller
-
     caller =
-      compile_time_application_and_file(file) ++
-        [module: module, function: form_fa(fun), line: line]
+      compile_time_application_and_file(caller) ++
+        case caller do
+          %{module: module, function: {fun, arity}, line: line} ->
+            [mfa: {module, fun, arity}, line: line]
+
+          _ ->
+            []
+        end
 
     {compile_metadata, quoted_metadata} =
       if Keyword.keyword?(metadata) do
         metadata = Keyword.merge(caller, metadata)
-        {metadata, metadata}
+        {Map.new(metadata), escape_metadata(metadata)}
       else
-        {[],
+        {%{},
          quote do
-           Keyword.merge(unquote(caller), unquote(metadata))
+           Enum.into(unquote(metadata), unquote(escape_metadata(caller)))
          end}
       end
 
@@ -804,19 +897,29 @@ defmodule Logger do
       no_log(data, quoted_metadata)
     else
       quote do
-        case Logger.__should_log__(unquote(level)) do
-          :error -> :ok
-          info -> Logger.__do_log__(info, unquote(data), unquote(quoted_metadata))
+        case Logger.__should_log__(unquote(level), __MODULE__) do
+          nil -> :ok
+          level -> Logger.__do_log__(level, unquote(data), unquote(quoted_metadata))
         end
       end
     end
   end
 
-  defp compile_time_application_and_file(file) do
+  defp escape_metadata(metadata) do
+    {_, metadata} =
+      Keyword.get_and_update(metadata, :mfa, fn
+        nil -> :pop
+        mfa -> {mfa, Macro.escape(mfa)}
+      end)
+
+    {:%{}, [], metadata}
+  end
+
+  defp compile_time_application_and_file(%{file: file}) do
     if app = Application.get_env(:logger, :compile_time_application) do
-      [application: app, file: Path.relative_to_cwd(file)]
+      [application: app, file: file |> Path.relative_to_cwd() |> String.to_charlist()]
     else
-      [file: file]
+      [file: String.to_charlist(file)]
     end
   end
 
@@ -826,10 +929,19 @@ defmodule Logger do
     Enum.any?(matching, fn filter ->
       Enum.all?(filter, fn
         {:level_lower_than, min_level} ->
-          Logger.Config.compare_levels(level, min_level) == :lt
+          compare_levels(level, min_level) == :lt
+
+        {:module, module} ->
+          match?({:ok, {^module, _, _}}, Map.fetch(compile_metadata, :mfa))
+
+        {:function, func} ->
+          case Map.fetch(compile_metadata, :mfa) do
+            {:ok, {_, f, a}} -> "#{f}/#{a}" == func
+            _ -> false
+          end
 
         {k, v} when is_atom(k) ->
-          Keyword.fetch(compile_metadata, k) == {:ok, v}
+          Map.fetch(compile_metadata, k) == {:ok, v}
 
         _ ->
           raise "expected :compile_time_purge_matching to be a list of keyword lists, " <>
@@ -838,11 +950,21 @@ defmodule Logger do
     end)
   end
 
-  # TODO: Deprecate compile_time_purge_level on v1.9
   defp maybe_log(level, data, metadata, caller) do
-    min_level = Application.get_env(:logger, :compile_time_purge_level, :debug)
+    min_level =
+      if env_level = Application.get_env(:logger, :compile_time_purge_level) do
+        IO.warn(
+          ":compile_time_purge_level option for the :logger application is deprecated, " <>
+            "use :compile_time_purge_matching instead",
+          Macro.Env.stacktrace(caller)
+        )
 
-    if Logger.Config.compare_levels(level, min_level) != :lt do
+        env_level
+      else
+        :debug
+      end
+
+    if compare_levels(level, min_level) != :lt do
       macro_log(level, data, metadata, caller)
     else
       no_log(data, metadata)
@@ -857,28 +979,4 @@ defmodule Logger do
       :ok
     end
   end
-
-  defp normalize_message(fun, metadata) when is_function(fun, 0) do
-    case fun.() do
-      {message, fun_metadata} -> {message, into_metadata(fun_metadata, metadata)}
-      :skip -> :skip
-      message -> {message, metadata}
-    end
-  end
-
-  defp normalize_message(message, metadata) do
-    {message, metadata}
-  end
-
-  defp truncate(data, n) when is_list(data) or is_binary(data), do: Logger.Utils.truncate(data, n)
-  defp truncate(data, n), do: Logger.Utils.truncate(to_string(data), n)
-
-  defp form_fa({name, arity}) do
-    Atom.to_string(name) <> "/" <> Integer.to_string(arity)
-  end
-
-  defp form_fa(nil), do: nil
-
-  defp notify(:sync, msg), do: :gen_event.sync_notify(Logger, msg)
-  defp notify(:async, msg), do: :gen_event.notify(Logger, msg)
 end
