@@ -63,10 +63,10 @@ defmodule Mix.Dep do
             system_env: []
 
   @type t :: %__MODULE__{
-          scm: module,
+          scm: Mix.SCM.t(),
           app: atom,
           requirement: String.t() | Regex.t() | nil,
-          status: atom,
+          status: {:ok, String.t() | nil} | atom | tuple,
           opts: keyword,
           top_level: boolean,
           manager: :rebar | :rebar3 | :mix | :make | nil,
@@ -124,7 +124,7 @@ defmodule Mix.Dep do
 
   defp load_and_cache(config, _top, bottom, _env, _target) do
     {_, deps} =
-      Mix.ProjectStack.read_cache({:cached_deps, bottom}) ||
+      Mix.State.read_cache({:cached_deps, bottom}) ||
         raise "cannot retrieve dependencies information because dependencies were not loaded. " <>
                 "Please invoke one of \"deps.loadpaths\", \"loadpaths\", or \"compile\" Mix task"
 
@@ -153,14 +153,14 @@ defmodule Mix.Dep do
   end
 
   defp read_cached_deps(project, env_target) do
-    case Mix.ProjectStack.read_cache({:cached_deps, project}) do
+    case Mix.State.read_cache({:cached_deps, project}) do
       {^env_target, deps} -> deps
       _ -> nil
     end
   end
 
   defp write_cached_deps(project, env_target, deps) do
-    Mix.ProjectStack.write_cache({:cached_deps, project}, {env_target, deps})
+    Mix.State.write_cache({:cached_deps, project}, {env_target, deps})
     deps
   end
 
@@ -170,7 +170,7 @@ defmodule Mix.Dep do
   def clear_cached() do
     if project = Mix.Project.get() do
       key = {:cached_deps, project}
-      Mix.ProjectStack.delete_cache(key)
+      Mix.State.delete_cache(key)
     end
   end
 
@@ -254,10 +254,13 @@ defmodule Mix.Dep do
     # mix.exs file can be different than the actual name and we
     # choose to respect the one in the mix.exs
     config =
-      Keyword.merge(Mix.Project.deps_config(), config)
+      Mix.Project.deps_config()
+      |> Keyword.merge(config)
       |> Keyword.put(:app_path, opts[:build])
       |> Keyword.put(:build_scm, scm)
 
+    config = Keyword.take(opts, [:inherit_parent_config_files]) ++ config
+
     env = opts[:env] || :prod
     old_env = Mix.env()
 
@@ -473,7 +476,7 @@ defmodule Mix.Dep do
   """
   def diverged?(%Mix.Dep{status: {:overridden, _}}), do: true
   def diverged?(%Mix.Dep{status: {:diverged, _}}), do: true
-  def diverged?(%Mix.Dep{status: {:divergedreq, _}}), do: true
+  def diverged?(%Mix.Dep{status: {:divergedreq, _, _}}), do: true
   def diverged?(%Mix.Dep{status: {:divergedonly, _}}), do: true
   def diverged?(%Mix.Dep{status: {:divergedtargets, _}}), do: true
   def diverged?(%Mix.Dep{}), do: false
