@@ -2,21 +2,17 @@ defmodule Mix.Tasks.Xref do
   use Mix.Task
 
   import Mix.Compilers.Elixir,
-    only: [read_manifest: 2, source: 0, source: 1, source: 2, module: 1]
+    only: [read_manifest: 1, source: 0, source: 1, source: 2, module: 1]
 
-  @shortdoc "Performs cross reference checks"
+  @shortdoc "Prints cross reference information"
   @recursive true
   @manifest "compile.elixir"
 
   @moduledoc """
-  Performs cross reference checks between modules.
+  Prints cross reference information between modules.
 
-  The unreachable and deprecated checks below happen every time
-  your project is compiled via `mix compile.xref`. See
-  `Mix.Tasks.Compile.Xref` for more information.
-
-  This task is automatically reenabled, so you can perform multiple
-  cross reference checks in the same Mix invocation.
+  This task is automatically reenabled, so you can print information
+  multiple times in the same Mix invocation.
 
   ## Xref modes
 
@@ -26,40 +22,11 @@ defmodule Mix.Tasks.Xref do
 
   All available modes are discussed below.
 
-  ### unreachable
-
-  Prints all unreachable "file:line: module.function/arity" entries:
-
-      mix xref unreachable
-
-  The "file:line" represents the file and line a call to an unknown
-  "module.function/arity" is made.
-
-  The option `--abort-if-any` can be used for the command to fail if
-  unreachable calls exist.
-
-  ## deprecated
-
-  Prints all deprecated "file:line: module.function/arity" entries:
-
-      mix xref deprecated
-
-  The "file:line" represents the file and line a call to a deprecated
-  "module.function/arity" is made. This operation does not show deprecated
-  local calls (a call to a deprecated function or macro in the same module)
-  nor calls to deprecated functionality in Elixir itself.
-
-  The option `--abort-if-any` can be used for the command to fail if
-  deprecated calls exist.
-
   ### callers CALLEE
 
-  Prints all callers of the given `CALLEE`, which can be one of: `Module`,
-  `Module.function`, or `Module.function/arity`. Examples:
+  Prints all callers of the given `MODULE`. Example:
 
       mix xref callers MyMod
-      mix xref callers MyMod.fun
-      mix xref callers MyMod.fun/3
 
   ### graph
 
@@ -130,14 +97,6 @@ defmodule Mix.Tasks.Xref do
 
     * `--no-elixir-version-check` - does not check the Elixir version from mix.exs
 
-  ## Configuration
-
-  All configuration for Xref should be placed under the key `:xref`.
-
-    * `:exclude` - a list of modules and `{module, function, arity}`
-      tuples to ignore when checking cross references. For example:
-      `[MissingModule, {MissingModule2, :missing_func, 2}]`
-
   """
 
   @switches [
@@ -167,33 +126,31 @@ defmodule Mix.Tasks.Xref do
     Mix.Task.reenable("xref")
 
     case args do
-      ["deprecated"] ->
-        deprecated(opts)
-
-      ["unreachable"] ->
-        unreachable(opts)
-
       ["callers", callee] ->
         callers(callee, opts)
 
       ["graph"] ->
         graph(opts)
 
+      # TODO: Remove on v2.0
+      ["deprecated"] ->
+        Mix.shell().error(
+          "The deprecated check has been moved to the compiler and has no effect now"
+        )
+
+      # TODO: Remove on v2.0
+      ["unreachable"] ->
+        Mix.shell().error(
+          "The unreachable check has been moved to the compiler and has no effect now"
+        )
+
       _ ->
         Mix.raise("xref doesn't support this command. For more information run \"mix help xref\"")
     end
   end
 
-  @doc false
-  def warnings(opts \\ []) do
-    source_warnings(excludes(), opts)
-    |> merge_entries(:all)
-    |> sort_entries()
-    |> print_warnings()
-  end
-
   @doc """
-  Returns a list of information of all the function calls in the project.
+  Returns a list of information of all the runtime function calls in the project.
 
   Each item in the list is a map with the following keys:
 
@@ -207,78 +164,129 @@ defmodule Mix.Tasks.Xref do
   information from. To get the function calls of each child in an umbrella,
   execute the function at the root of each individual application.
   """
+  # TODO: Remove on v2.0
+  @doc deprecated: "Use compilation tracers described in the Code module"
   @spec calls(keyword()) :: [
           %{
             callee: {module(), atom(), arity()},
-            caller_module: module(),
             line: integer(),
             file: String.t()
           }
         ]
   def calls(opts \\ []) do
-    module_sources =
-      for manifest <- manifests(opts),
-          manifest_data = read_manifest(manifest, ""),
-          module(module: module, sources: sources) <- manifest_data,
-          source <- sources,
-          source = Enum.find(manifest_data, &match?(source(source: ^source), &1)),
-          do: {module, source}
+    for manifest <- manifests(opts),
+        source(source: source, modules: modules) <- read_manifest(manifest) |> elem(1),
+        module <- modules,
+        call <- collect_calls(source, module),
+        do: call
+  end
 
-    Enum.flat_map(module_sources, fn {caller_module, source} ->
-      source(
-        runtime_dispatches: runtime_nested,
-        compile_dispatches: compile_nested,
-        source: rel_file
-      ) = source
+  defp collect_calls(source, module) do
+    with [_ | _] = path <- :code.which(module),
+         {:ok, {_, [debug_info: debug_info]}} <- :beam_lib.chunks(path, [:debug_info]),
+         {:debug_info_v1, backend, data} <- debug_info,
+         {:ok, %{definitions: defs}} <- backend.debug_info(:elixir_v1, module, data, []),
+         do: walk_definitions(module, source, defs)
+  end
 
-      runtime_function_calls =
-        dispatches_to_function_calls(caller_module, rel_file, runtime_nested)
+  defp walk_definitions(module, file, definitions) do
+    state = %{
+      file: file,
+      module: module,
+      calls: []
+    }
 
-      compile_function_calls =
-        dispatches_to_function_calls(caller_module, rel_file, compile_nested)
+    state = Enum.reduce(definitions, state, &walk_definition/2)
+    state.calls
+  end
 
-      runtime_function_calls ++ compile_function_calls
+  defp walk_definition({_function, _kind, meta, clauses}, state) do
+    with_file_meta(state, meta, fn state ->
+      Enum.reduce(clauses, state, &walk_clause/2)
     end)
   end
 
-  defp dispatches_to_function_calls(caller_module, file, dispatches) do
-    for {module, function_calls} <- dispatches,
-        {{function, arity}, lines} <- function_calls,
-        line <- lines do
-      %{
-        callee: {module, function, arity},
-        file: file,
-        line: line,
-        caller_module: caller_module
-      }
+  defp with_file_meta(%{file: original_file} = state, meta, fun) do
+    case Keyword.fetch(meta, :file) do
+      {:ok, {meta_file, _}} ->
+        state = fun.(%{state | file: meta_file})
+        %{state | file: original_file}
+
+      :error ->
+        fun.(state)
     end
   end
 
+  defp walk_clause({_meta, args, _guards, body}, state) do
+    state = walk_expr(args, state)
+    walk_expr(body, state)
+  end
+
+  # &Mod.fun/arity
+  defp walk_expr({:&, meta, [{:/, _, [{{:., _, [module, fun]}, _, []}, arity]}]}, state)
+       when is_atom(module) and is_atom(fun) do
+    add_call(module, fun, arity, meta, state)
+  end
+
+  # Mod.fun(...)
+  defp walk_expr({{:., _, [module, fun]}, meta, args}, state)
+       when is_atom(module) and is_atom(fun) do
+    state = add_call(module, fun, length(args), meta, state)
+    walk_expr(args, state)
+  end
+
+  # %Module{...}
+  defp walk_expr({:%, meta, [module, {:%{}, _meta, args}]}, state)
+       when is_atom(module) and is_list(args) do
+    state = add_call(module, :__struct__, 0, meta, state)
+    walk_expr(args, state)
+  end
+
+  # Function call
+  defp walk_expr({left, _meta, right}, state) when is_list(right) do
+    state = walk_expr(right, state)
+    walk_expr(left, state)
+  end
+
+  # {x, y}
+  defp walk_expr({left, right}, state) do
+    state = walk_expr(right, state)
+    walk_expr(left, state)
+  end
+
+  # [...]
+  defp walk_expr(list, state) when is_list(list) do
+    Enum.reduce(list, state, &walk_expr/2)
+  end
+
+  defp walk_expr(_other, state) do
+    state
+  end
+
+  defp add_call(module, fun, arity, meta, state) do
+    call = %{
+      callee: {module, fun, arity},
+      caller_module: state.module,
+      file: state.file,
+      line: meta[:line]
+    }
+
+    %{state | calls: [call | state.calls]}
+  end
+
   ## Modes
 
-  defp unreachable(opts) do
-    source_warnings(excludes(), opts)
-    |> merge_entries(&(&1 in [:unknown_module, :unknown_function]))
-    |> sort_entries()
-    |> print_mfas()
-    |> return_ok_error_or_abort(:unreachable, opts)
-  end
-
-  defp deprecated(opts) do
-    source_warnings(excludes(), opts)
-    |> merge_entries(&(&1 == :deprecated))
-    |> sort_entries()
-    |> print_mfas()
-    |> return_ok_error_or_abort(:deprecated, opts)
-  end
-
   defp callers(callee, opts) do
-    callee
-    |> filter_for_callee()
-    |> source_callers(opts)
-    |> merge_entries(:all)
-    |> sort_entries()
-    |> print_calls()
+    module = parse_callee(callee)
+
+    file_callers =
+      for source <- sources(opts),
+          reference = reference(module, source),
+          do: {source(source, :source), reference}
+
+    for {file, type} <- Enum.sort(file_callers) do
+      Mix.shell().info([file, " (", type, ")"])
+    end
 
     :ok
   end
@@ -289,259 +297,21 @@ defmodule Mix.Tasks.Xref do
     :ok
   end
 
-  ## Warnings (unreachable + deprecated)
-
-  defp source_warnings(excludes, opts) do
-    sources = sources(opts)
-
-    exports_and_deprecated =
-      Enum.reduce(sources, %{}, fn source, acc ->
-        runtime = source(source, :runtime_dispatches)
-        compile = source(source, :compile_dispatches)
-
-        acc = Enum.reduce(runtime, acc, &load_exports_and_deprecated_into_acc/2)
-        acc = Enum.reduce(compile, acc, &load_exports_and_deprecated_into_acc/2)
-        acc
-      end)
-
-    runtime_warnings =
-      for source <- sources,
-          {module, func_arity_locations} <- source(source, :runtime_dispatches),
-          {exports, deprecated} = exports_and_deprecated[module],
-          {{func, arity}, locations} <- func_arity_locations,
-          not excluded?(excludes, module, func, arity),
-          warning =
-            unreachable_mfa(exports, module, func, arity) ||
-              deprecated_mfa(deprecated, module, func, arity),
-          do: {warning, absolute_locations(locations, source(source, :source))}
-
-    compile_warnings =
-      for source <- sources,
-          {module, func_arity_locations} <- source(source, :compile_dispatches),
-          {_, deprecated} = exports_and_deprecated[module],
-          {{func, arity}, locations} <- func_arity_locations,
-          not excluded?(excludes, module, func, arity),
-          warning = deprecated_mfa(deprecated, module, func, arity),
-          do: {warning, absolute_locations(locations, source(source, :source))}
-
-    runtime_warnings ++ compile_warnings
-  end
-
-  defp load_exports_and_deprecated_into_acc({module, _}, acc) do
-    case acc do
-      %{^module => _} -> acc
-      _ -> Map.put(acc, module, load_exports_and_deprecated(module))
-    end
-  end
-
-  defp load_exports_and_deprecated(module) do
-    if :code.is_loaded(module) do
-      # If the module is loaded, we will use the faster function_exported?/3
-      # check for exports and __info__/1 for deprecated
-      if function_exported?(module, :__info__, 1) do
-        {module, load_deprecated_from_module(module)}
-      else
-        {module, []}
-      end
-    else
-      # Otherwise we get the information from :beam_lib to avoid loading modules
-      with [_ | _] = file <- :code.which(module),
-           {:ok, {^module, [{:exports, exports}, {'ExDp', deprecated}]}} <-
-             :beam_lib.chunks(file, [:exports, 'ExDp'], [:allow_missing_chunks]) do
-        {exports, load_deprecated_from_chunk(deprecated)}
-      else
-        _ -> {:unknown_module, []}
-      end
-    end
-  end
-
-  defp load_deprecated_from_chunk(chunk) do
-    if is_binary(chunk) do
-      {:elixir_deprecated_v1, deprecated} = :erlang.binary_to_term(chunk)
-      deprecated
-    else
-      []
-    end
-  end
-
-  defp load_deprecated_from_module(module) do
-    try do
-      module.__info__(:deprecated)
-    rescue
-      _ -> []
-    end
-  end
-
-  defp unreachable_mfa(exports, module, func, arity) do
-    cond do
-      skip_unreachable?(module, func, arity) ->
-        nil
-
-      exports == :unknown_module ->
-        {:unknown_module, module, func, arity, nil}
-
-      is_atom(exports) and not function_exported?(module, func, arity) ->
-        {:unknown_function, module, func, arity, nil}
-
-      is_list(exports) and {func, arity} not in exports ->
-        {:unknown_function, module, func, arity, exports}
-
-      true ->
-        nil
-    end
-  end
-
-  defp deprecated_mfa(deprecated, module, func, arity) do
-    case List.keyfind(deprecated, {func, arity}, 0) do
-      {_, reason} ->
-        {:deprecated, module, func, arity, reason}
-
-      nil ->
-        nil
-    end
-  end
-
-  @protocol_built_ins for {_, type} <- Protocol.__built_in__(), do: type
-
-  defp skip_unreachable?(:erlang, func, 2) when func in [:andalso, :orelse] do
-    true
-  end
-
-  defp skip_unreachable?(module, :__impl__, 1) do
-    {maybe_protocol, maybe_built_in} = module |> Module.split() |> Enum.split(-1)
-    maybe_protocol = Module.concat(maybe_protocol)
-    maybe_built_in = Module.concat(maybe_built_in)
-
-    maybe_built_in in @protocol_built_ins and Code.ensure_loaded?(maybe_protocol) and
-      function_exported?(maybe_protocol, :__protocol__, 1)
-  end
-
-  defp skip_unreachable?(_, _, _) do
-    false
-  end
-
-  defp excludes() do
-    Mix.Project.config()
-    |> Keyword.get(:xref, [])
-    |> Keyword.get(:exclude, [])
-    |> MapSet.new()
-  end
-
-  defp excluded?(excludes, module, func, arity) do
-    MapSet.member?(excludes, module) or MapSet.member?(excludes, {module, func, arity})
-  end
-
-  ## Print mfas (unreachable / deprecated)
-
-  defp print_mfas(entries) do
-    Enum.each(entries, &print_mfa/1)
-    entries
-  end
-
-  defp print_mfa({{_, module, function, arity, _}, locations}) do
-    shell = Mix.shell()
-
-    for {file, line} <- locations do
-      shell.info([
-        Exception.format_file_line(file, line, " "),
-        Exception.format_mfa(module, function, arity)
-      ])
-    end
-  end
-
-  ## Print warnings
-
-  defp print_warnings(entries) do
-    prefix = IO.ANSI.format([:yellow, "warning: "])
-
-    Enum.map(entries, fn {type, locations} ->
-      message = warning_message(type)
-      IO.write(:stderr, [prefix, message, ?\n, format_locations(locations), ?\n])
-      {message, locations}
-    end)
-  end
-
-  defp warning_message({:unknown_function, module, function, arity, exports}) do
-    [
-      "function ",
-      Exception.format_mfa(module, function, arity),
-      " is undefined or private",
-      UndefinedFunctionError.hint_for_loaded_module(module, function, arity, exports)
-    ]
-  end
-
-  defp warning_message({:unknown_module, module, function, arity, _}) do
-    [
-      "function ",
-      Exception.format_mfa(module, function, arity),
-      " is undefined (module #{inspect(module)} is not available)"
-    ]
-  end
-
-  defp warning_message({:deprecated, module, function, arity, reason}) do
-    [
-      Exception.format_mfa(module, function, arity),
-      " is deprecated. ",
-      reason
-    ]
-  end
-
-  defp format_locations([location]) do
-    format_location(location)
-  end
-
-  defp format_locations(locations) do
-    [
-      "Found at #{length(locations)} locations:\n",
-      Enum.map(locations, &format_location/1)
-    ]
-  end
-
-  defp format_location({file, line}) do
-    ["  ", file, ?:, Integer.to_string(line), ?\n]
-  end
-
   ## Callers
 
-  defp source_callers(filter, opts) do
-    for source <- sources(opts),
-        file = source(source, :source),
-        {module, func_arity_locations} <- dispatches(source),
-        {{func, arity}, locations} <- func_arity_locations,
-        filter.({module, func, arity}),
-        do: {{module, func, arity}, absolute_locations(locations, file)}
-  end
-
-  defp print_calls(calls) do
-    Enum.each(calls, &print_call/1)
-    calls
-  end
-
-  defp print_call({{module, func, arity}, locations}) do
-    shell = Mix.shell()
-
-    for {file, line} <- locations do
-      shell.info([
-        Exception.format_file_line(file, line, " "),
-        Exception.format_mfa(module, func, arity)
-      ])
+  defp parse_callee(callee) do
+    case Mix.Utils.parse_mfa(callee) do
+      {:ok, [module]} -> module
+      _ -> Mix.raise("xref callers MODULE expects a MODULE, got: " <> callee)
     end
   end
 
-  defp filter_for_callee(callee) do
-    case Mix.Utils.parse_mfa(callee) do
-      {:ok, mfa_list} ->
-        mfa_list_length = length(mfa_list)
-
-        fn {module, function, arity} ->
-          mfa_list == Enum.take([module, function, arity], mfa_list_length)
-        end
-
-      :error ->
-        Mix.raise(
-          "xref callers CALLEE expects Module, Module.function, or Module.function/arity, " <>
-            "got: " <> callee
-        )
+  defp reference(module, source) do
+    cond do
+      module in source(source, :compile_references) -> "compile"
+      module in source(source, :struct_references) -> "struct"
+      module in source(source, :runtime_references) -> "runtime"
+      true -> nil
     end
   end
 
@@ -564,10 +334,10 @@ defmodule Mix.Tasks.Xref do
 
     module_sources =
       for manifest_path <- manifests(opts),
-          manifest_data = read_manifest(manifest_path, ""),
-          module(module: module, sources: sources) <- manifest_data,
+          {manifest_modules, manifest_sources} = read_manifest(manifest_path),
+          module(module: module, sources: sources) <- manifest_modules,
           source <- sources,
-          source = Enum.find(manifest_data, &match?(source(source: ^source), &1)),
+          source = Enum.find(manifest_sources, &match?(source(source: ^source), &1)),
           do: {module, source}
 
     all_modules = MapSet.new(module_sources, &elem(&1, 0))
@@ -743,7 +513,7 @@ defmodule Mix.Tasks.Xref do
 
   defp sources(opts) do
     for manifest <- manifests(opts),
-        source() = source <- read_manifest(manifest, ""),
+        source() = source <- read_manifest(manifest) |> elem(1),
         do: source
   end
 
@@ -759,40 +529,4 @@ defmodule Mix.Tasks.Xref do
 
     [Path.join(Mix.Project.manifest_path(), @manifest) | siblings]
   end
-
-  defp merge_entries(entries, filter) do
-    Enum.reduce(entries, %{}, fn {type, locations}, merged_entries ->
-      if filter == :all or filter.(elem(type, 0)) do
-        locations = MapSet.new(locations)
-        Map.update(merged_entries, type, locations, &MapSet.union(&1, locations))
-      else
-        merged_entries
-      end
-    end)
-  end
-
-  defp sort_entries(entries) do
-    entries
-    |> Enum.map(fn {type, locations} -> {type, Enum.sort(locations)} end)
-    |> Enum.sort()
-  end
-
-  defp absolute_locations(locations, base) do
-    Enum.map(locations, &absolute_location(&1, base))
-  end
-
-  defp absolute_location({_, _} = location, _), do: location
-  defp absolute_location(line, base), do: {base, line}
-
-  defp dispatches(source) do
-    source(source, :runtime_dispatches) ++ source(source, :compile_dispatches)
-  end
-
-  defp return_ok_error_or_abort(entries, kind, opts) do
-    cond do
-      entries == [] -> :ok
-      opts[:abort_if_any] -> Mix.raise("mix xref #{kind} failed: #{kind} calls were found")
-      true -> :error
-    end
-  end
 end
