@@ -1,4 +1,4 @@
-use Mix.Config
+import Config
 
 config :release_test, :static, :was_set
 config :release_test, :encoding, {:"£", "£", '£'}
